"""Unit tests for the web API basic auth backend."""

from __future__ import unicode_literals

import base64
import logging

from django.contrib.auth.models import AnonymousUser, User
from django.contrib.sessions.middleware import SessionMiddleware
from django.test.client import RequestFactory
from kgb import SpyAgency

from djblets.testing.testcases import TestCase
from djblets.webapi.auth.backends.basic import WebAPIBasicAuthBackend


class WebAPIBasicAuthBackendTests(SpyAgency, TestCase):
    """Unit tests for the WebAPIBasicAuthBackend."""

    def setUp(self):
        super(WebAPIBasicAuthBackendTests, self).setUp()

        self.basic_auth_backend = WebAPIBasicAuthBackend()
        self.request = RequestFactory().get('/')
        SessionMiddleware().process_request(self.request)
        self.request.user = AnonymousUser()
        self.user = User.objects.create_user(
            username='testuser',
            email='testcase@test.com',
            password='testpassword')

    def test_get_credentials_no_basic_realm(self):
        """Testing Basic Auth get_credentials with non-basic realm"""
        self.request.META['HTTP_AUTHORIZATION'] = 'Token tokenHere'
        result = self.basic_auth_backend.get_credentials(self.request)
        self.assertIsNone(result)

    def test_get_credentials_valid_credentials(self):
        """Testing Basic Auth get_credentials with credentials"""
        encoded_credentials = base64.b64encode(b'my-username:my-password')
        header = 'Basic %s' % encoded_credentials.decode('utf-8')
        self.request.META['HTTP_AUTHORIZATION'] = header

        result = self.basic_auth_backend.get_credentials(self.request)

        self.assertEqual(result, {
            'username': 'my-username',
            'password': 'my-password',
        })

    def test_get_credentials_malformed_credentials(self):
        """Testing Basic Auth get_credentials with malformed credentials"""
        self.request.META['HTTP_AUTHORIZATION'] = (
            'Basic %s'
            % base64.b64encode(b'Some malfomred credentials').decode('utf-8')
        )

        logger = logging.getLogger('djblets.webapi.auth.backends.basic')
        self.spy_on(logger.warning)

        result = self.basic_auth_backend.get_credentials(self.request)
        warning_message = logger.warning.spy.last_call.args[0]

        self.assertIsNone(result)
        self.assertTrue(warning_message.startswith(
                        'Failed to parse HTTP_AUTHORIZATION header'))

    def test_authenticate_valid_credentials(self):
        """Testing Basic Auth authenicate with valid credentials"""
        header = ('Basic %s'
                  % base64.b64encode(b'testuser:testpassword').decode('utf-8'))
        self.request.META['HTTP_AUTHORIZATION'] = header

        result = self.basic_auth_backend.authenticate(self.request)

        self.assertEqual(result, (True, None, None))

    def test_authenticate_wrong_header(self):
        """Testing Basic Auth authenicate with wrong header"""
        header = ('Token %s'
                  % base64.b64encode(b'testuser:testpassword').decode('utf-8'))
        self.request.META['HTTP_AUTHORIZATION'] = header

        result = self.basic_auth_backend.authenticate(self.request)

        self.assertIsNone(result)

    def test_authenticate_wrong_password(self):
        """Testing Basic Auth authenicate with invalid credentials"""
        header = (
            'Basic %s'
            % base64.b64encode(b'testuser:wrongpassword').decode('utf-8'))
        self.request.META['HTTP_AUTHORIZATION'] = header

        result = self.basic_auth_backend.authenticate(self.request)

        self.assertEqual(result, (False, None, None))

    def test_login_with_credentials_valid_credentials(self):
        """Testing Basic Auth login_with_credentials with valid credentials"""
        encoded_credentials = base64.b64encode(b'testuser:testpassword')
        header = 'Basic %s' % encoded_credentials.decode('utf-8')
        self.request.META['HTTP_AUTHORIZATION'] = header

        result = self.basic_auth_backend.login_with_credentials(
            self.request,
            username='testuser',
            password='testpassword')

        self.assertEqual(result, (True, None, None))

    def test_login_bypass_authentication(self):
        """Testing Basic Auth login_with_credentials with currently logged in
        user
        """
        encoded_credentials = base64.b64encode(b'testuser:testpassword')
        header = 'Basic %s' % encoded_credentials.decode('utf-8')
        self.request.META['HTTP_AUTHORIZATION'] = header
        self.request.user = self.user

        result = self.basic_auth_backend.login_with_credentials(
            self.request,
            username='testuser',
            password='testpassword')

        self.assertEqual(result, (True, None, None))

    def test_login_with_credentials_incorrect_pass(self):
        """Testing Basic Auth login_with_credentials with incorrect password"""
        encoded_credentials = base64.b64encode(b'testuser:wrongpassword')
        header = 'Basic %s' % encoded_credentials.decode('utf-8')
        self.request.META['HTTP_AUTHORIZATION'] = header

        result = self.basic_auth_backend.login_with_credentials(
            self.request,
            username='testuser',
            password='wrongpassword')

        self.assertEqual(result, (False, None, None))

    def test_login_with_credentials_incorrect_user(self):
        """Testing Basic Auth login_with_credentials with invalid user"""
        encoded_credentials = base64.b64encode(b'wronguser:testpassword')
        header = 'Basic %s' % encoded_credentials.decode('utf-8')
        self.request.META['HTTP_AUTHORIZATION'] = header
        self.request.user = self.user

        result = self.basic_auth_backend.login_with_credentials(
            self.request,
            username='wronguser',
            password='testpassword')

        self.assertEqual(result, (False, None, None))

    def test_validate_credentials_valid(self):
        """Testing Basic Auth validate_credentials with valid credentials"""
        self.request.user = self.user
        result = self.basic_auth_backend.validate_credentials(
            self.request,
            username='testuser',
            password='testpassword')

        self.assertEqual(result, (True, None, None))

    def test_validate_credentials_invalid_user(self):
        """Testing Basic Auth validate_credentials with invalid user"""
        result = self.basic_auth_backend.validate_credentials(
            self.request,
            username='testuser',
            password='testpassword')

        self.assertIsNone(result)

    def test_validate_credentials_invalid_credentials(self):
        """Testing Basic Auth validate_credentials with invalid credentials"""
        self.request.user = self.user
        result = self.basic_auth_backend.validate_credentials(
            self.request,
            username='differentuser',
            password='testpassword')

        self.assertIsNone(result)

    def test_clean_credentials_for_display_removes_credentials(self):
        """Testing Basic Auth clean_credentials_for_display"""
        credentials = {
            'api': 'has exact match',
            'oauth2_token': 'has partial match',
            'apikey2': 'no match as word but contains match sub string',
            'secre': 'no match, similar only',
        }
        removed_credential = '************'
        clean_credentials = \
            self.basic_auth_backend.clean_credentials_for_display(credentials)

        self.assertEqual(clean_credentials['api'], removed_credential)
        self.assertEqual(clean_credentials['oauth2_token'], removed_credential)
        self.assertEqual(clean_credentials['apikey2'], removed_credential)
        self.assertEqual(clean_credentials['secre'], credentials['secre'])



from __future__ import unicode_literals

from django.utils import six
from kgb import SpyAgency

from djblets.registries.errors import (AlreadyRegisteredError,
                                       ItemLookupError,
                                       RegistrationError)
from djblets.registries.registry import Registry, OrderedRegistry, UNREGISTER
from djblets.registries.signals import registry_populating
from djblets.testing.testcases import TestCase


class Item(object):
    """An object with arbitrary attributes."""

    def __init__(self, **attrs):
        """Initialize the item.

        Args:
            **attrs (dict):
                Attribute name and value pairs to set on the item.
        """
        self._attrs = set(attr_name for attr_name in attrs)

        for attr_name, attr_value in six.iteritems(attrs):
            setattr(self, attr_name, attr_value)

    def __repr__(self):
        attrs = (
            '%s=%r' % (attr_name, getattr(self, attr_name))
            for attr_name in self._attrs
        )
        return '<Item(%s)>' % ', '.join(attrs)

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return (type(self) is type(other) and
                self._attrs == other._attrs and
                all(getattr(self, attr_name) == getattr(other, attr_name)
                    for attr_name in self._attrs))


class RegistryTests(SpyAgency, TestCase):
    """Tests for djblets.registries.Registry."""

    def test_empty_by_default(self):
        """Testing Registry instances are created empty"""
        r = Registry()

        self.assertEqual(len(r), 0)
        self.assertEqual(set(r), set())

    def test_get_invalid_attribute(self):
        """Testing Registry.get_item with invalid attributes"""
        r = Registry()

        with self.assertRaises(ItemLookupError):
            r.get('foo', 'bar')

    def test_register_item(self):
        """Testing Registry.register_item"""
        r = Registry()
        items = [1, 2, 3]

        for item in items:
            r.register(item)

        self.assertEqual(set(r), set(items))

    def test_unregister_item(self):
        """Testing Registry.unregister_item"""
        r = Registry()
        items = [1, 2, 3]

        for item in items:
            r.register(item)

        for item in items:
            r.unregister(item)

        self.assertEqual(set(r), set())

    def test_unregister_removes_attr_lookups(self):
        """Testing Registry.unregister removes lookup entries"""
        class TestRegistry(Registry):
            lookup_attrs = ('id',)

        r = TestRegistry()
        items = [Item(id=0), Item(id=1), Item(id=2)]

        for item in items:
            r.register(item)

        for item in items:
            r.unregister(item)

            with self.assertRaises(ItemLookupError):
                r.get('id', item.id)

        self.assertEqual(len(r), 0)

    def test_population_on_register(self):
        """Testing Registry.register_item triggers population before
        registration
        """
        original_item = Item(id=0)
        duplicate_item = Item(id=0, fake=True)

        class TestRegistry(Registry):
            lookup_attrs = ('id',)

            def get_defaults(self):
                yield original_item

        self.spy_on(registry_populating.send)

        r = TestRegistry()

        with self.assertRaises(AlreadyRegisteredError):
            r.register(duplicate_item)

        self.assertIs(r.get('id', original_item.id),
                      original_item)
        self.assertTrue(registry_populating.send.called_with(
            sender=TestRegistry,
            registry=r))

    def test_population_on_unregister(self):
        """Testing Registry.unregister_item triggers population before
        unregistration.
        """
        item = Item(id=0)

        class TestRegistry(Registry):
            lookup_attrs = ('id',)

            def get_defaults(self):
                yield item

        self.spy_on(registry_populating.send)

        r = TestRegistry()
        r.unregister(item)
        self.assertEqual(len(r), 0)

        self.assertTrue(registry_populating.send.called_with(
            sender=TestRegistry,
            registry=r))

    def test_registering_duplicate(self):
        """Testing Registry.register_item with duplicate items"""
        r = Registry()
        r.register(1)

        with self.assertRaises(AlreadyRegisteredError):
            r.register(1)

    def test_registering_duplicate_attributes(self):
        """Testing Registry.register_item with items that have identical
        attributes
        """
        class TestRegistry(Registry):
            lookup_attrs = ('id',)

        r = TestRegistry()
        r.register(Item(id=0, name='foo'))

        with self.assertRaises(AlreadyRegisteredError):
            r.register(Item(id=0, name='bar'))

    def test_register_missing_attributes(self):
        """Testing Registry.register_item with items that have missing
        attributes
        """
        class TestRegistry(Registry):
            lookup_attrs = ('id',)

        r = TestRegistry()

        with self.assertRaises(RegistrationError):
            r.register(Item())

    def test_contains(self):
        """Testing Registry.__contains__"""
        r = Registry()
        r.register(1)

        self.assertTrue(1 in r)
        self.assertFalse(2 in r)

    def test_error_override(self):
        """Testing Registry error formatting strings"""
        class TestRegistry(Registry):
            errors = {
                UNREGISTER: 'The foo "%(item)s" is unregistered.'
            }

        r = TestRegistry()

        with self.assertRaisesMessage(ItemLookupError,
                                      'The foo "1" is unregistered.'):
            r.unregister(1)


class OrderedRegistryTests(TestCase):
    """Tests for djblets.registries.registry.OrderedRegistry."""

    class TestRegistry(OrderedRegistry):
        def get_defaults(self):
            return [1, 2, 3]

    def setUp(self):
        self.registry = OrderedRegistryTests.TestRegistry()

    def test_iteration_order(self):
        """Testing OrderedRegistry iteration order"""
        self.registry.register(4)
        self.assertListEqual(list(self.registry),
                             [1, 2, 3, 4])

    def test_getitem(self):
        """Testing OrderedRegistry.__getitem__"""
        self.assertTrue(self.registry[0], 1)
        self.assertTrue(self.registry[1], 2)
        self.assertTrue(self.registry[2], 3)

    def test_getitem_negative_indices(self):
        """Testing OrderedRegistry.__getitem__ with negative indices"""
        self.assertTrue(self.registry[-1], 3)
        self.assertTrue(self.registry[-2], 2)
        self.assertTrue(self.registry[-3], 1)

    def test_getitem_invalid_index(self):
        """Testing OrderedRegistry.__getitem__ with an invalid index"""
        with self.assertRaises(TypeError):
            self.registry['foo']

    def test_getitem_out_of_range(self):
        """Testing OrderedRegistry.__getitem__ with an out of range index"""
        with self.assertRaises(IndexError):
            self.registry[1000]



from __future__ import unicode_literals

from django_evolution.mutations import RenameAppLabel


MUTATIONS = [
    RenameAppLabel('siteconfig', 'djblets_siteconfig',
                   legacy_app_label='siteconfig'),
]



"""Unit tests for djblets.privacy.consent.forms.ConsentFormMixin."""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.forms import Form

from djblets.privacy.consent import (BaseConsentRequirement, Consent,
                                     get_consent_requirements_registry)
from djblets.privacy.consent.forms import ConsentFormMixin
from djblets.privacy.tests.testcases import ConsentTestCase


class MyForm(ConsentFormMixin, Form):
    def __init__(self, user=None, *args, **kwargs):
        self.user = user

        super(MyForm, self).__init__(*args, **kwargs)

    def get_consent_user(self):
        return self.user

    def get_consent_source(self):
        return 'https://example.com/consent/'

    def get_extra_consent_data(self):
        return {
            'test': True,
        }


class MyConsentRequirement1(BaseConsentRequirement):
    requirement_id = 'my-requirement-1'
    name = 'My Requirement 1'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'
    icons = {
        '1x': '/static/consent.png',
        '2x': '/static/consent@2x.png',
    }


class MyConsentRequirement2(BaseConsentRequirement):
    requirement_id = 'my-requirement-2'
    name = 'My Requirement 2'
    summary = 'We would also like this'
    intent_description = 'We need this for dancing.'
    data_use_description = 'Dancing all the things.'


class ConsentFormMixinTests(ConsentTestCase):
    """Unit tests for ConsentFormMixinTests."""

    def setUp(self):
        super(ConsentFormMixinTests, self).setUp()

        self.registry = get_consent_requirements_registry()

        self.consent_requirement_1 = MyConsentRequirement1()
        self.registry.register(self.consent_requirement_1)

        self.consent_requirement_2 = MyConsentRequirement2()
        self.registry.register(self.consent_requirement_2)

        self.user = User.objects.create(username='test-user')

    def test_init(self):
        """Testing ConsentFormMixin.__init__ defines field"""
        form = MyForm()
        self.assertIn('consent', form.fields)

        field = form.fields['consent']
        self.assertEqual(
            field.consent_requirements,
            [self.consent_requirement_1, self.consent_requirement_2])

        for subfield in field.fields:
            self.assertEqual(subfield.consent_source,
                             'https://example.com/consent/')
            self.assertEqual(
                subfield.extra_consent_data,
                {
                    'test': True,
                })

    def test_save_consent(self):
        """Testing ConsentFormMixin.save_consent"""
        form = MyForm(data={
            'consent_my-requirement-1_choice': 'allow',
            'consent_my-requirement-2_choice': 'block',
        })
        self.assertTrue(form.is_valid())

        self.assertEqual(self.consent_requirement_1.get_consent(self.user),
                         Consent.UNSET)
        self.assertEqual(self.consent_requirement_2.get_consent(self.user),
                         Consent.UNSET)

        form.save_consent(self.user)

        self.assertEqual(self.consent_requirement_1.get_consent(self.user),
                         Consent.GRANTED)
        self.assertEqual(self.consent_requirement_2.get_consent(self.user),
                         Consent.DENIED)



"""Error classes for conditions."""

from __future__ import unicode_literals

from djblets.registries.errors import AlreadyRegisteredError, ItemLookupError


class ConditionChoiceConflictError(AlreadyRegisteredError):
    """A condition choice conflicts with another registered choice."""


class ConditionChoiceNotFoundError(ItemLookupError):
    """A condition choice was not found."""

    def __init__(self, message, condition_index=None, choice_id=None):
        """Initialize the error.

        Args:
            message (unicode):
                The error message.

            condition_index (int, optional):
                The index of the condition this error applies to within
                the condition set.

            choice_id (unicode, optional):
                The ID of the choice that could not be found.
        """
        super(ConditionChoiceNotFoundError, self).__init__(message)

        self.condition_index = condition_index
        self.choice_id = choice_id


class ConditionOperatorConflictError(AlreadyRegisteredError):
    """A condition operator conflicts with another registered operator."""


class ConditionOperatorNotFoundError(ItemLookupError):
    """A condition operator was not found."""

    def __init__(self, message, condition_index=None, operator_id=None):
        """Initialize the error.

        Args:
            message (unicode):
                The error message.

            condition_index (int, optional):
                The index of the condition this error applies to within
                the condition set.

            choice_id (unicode, optional):
                The ID of the operator that could not be found.
        """
        super(ConditionOperatorNotFoundError, self).__init__(message)

        self.condition_index = condition_index
        self.operator_id = operator_id


class InvalidConditionModeError(ValueError):
    """The condition mode provided was invalid."""


class InvalidConditionValueError(ValueError):
    """The condition value provided was invalid."""

    def __init__(self, message, code=None, condition_index=None):
        """Initialize the error.

        Args:
            message (unicode):
                The error message.

            code (unicode, optional):
                The error code. This will generally correspond to a form
                validation error code.

            condition_index (int, optional):
                The index of the condition this error applies to within
                the condition set.
        """
        super(InvalidConditionValueError, self).__init__(message)

        self.code = code
        self.condition_index = condition_index



"""Unit tests for djblets.forms.widgets.CopyableTextInput."""

from __future__ import unicode_literals

from django import forms

from djblets.forms.widgets import CopyableTextInput
from djblets.testing.testcases import TestCase


class CopyableTextInputTests(TestCase):
    """Unit tests for djblets.forms.widgets.CopyableTextInput."""

    def test_render(self):
        """Testing CopyableTextInput.render"""
        field = forms.CharField(widget=CopyableTextInput())
        rendered = field.widget.render(
            name='my_field',
            value='test',
            attrs={
                'id': 'id_my_field',
            })

        self.assertIn('class="copyable-text-input-link"', rendered)
        self.assertIn('data-field-id="id_my_field"', rendered)
        self.assertIn('title="Copy to clipboard"', rendered)



"""Unit tests for djblets.extensions.hooks.BaseRegistryMultiItemsHook."""

from __future__ import unicode_literals

from django.utils import six

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import (ExtensionHookPoint,
                                      BaseRegistryMultiItemHook)
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.registries.errors import AlreadyRegisteredError
from djblets.registries.registry import Registry
from djblets.testing.testcases import TestCase


class DummyRegistry(Registry):
    lookup_attrs = ('foo_id',)


class DummyItem(object):
    def __init__(self, foo_id):
        self.foo_id = foo_id


class BaseRegistryMultiItemHookTests(ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.BaseRegistryMultiItemsHook."""

    def setUp(self):
        super(BaseRegistryMultiItemHookTests, self).setUp()

        self.registry = DummyRegistry()

        @six.add_metaclass(ExtensionHookPoint)
        class DummyRegistryHook(BaseRegistryMultiItemHook):
            registry = self.registry

        class TestExtension(Extension):
            pass

        self.hook_cls = DummyRegistryHook

        self.extension = self.setup_extension(TestExtension)

    def test_initialize(self):
        """Testing BaseRegistryMultiItemHook.initialize"""
        self.assertEqual(list(self.registry), [])

        item1 = DummyItem(123)
        item2 = DummyItem(456)
        self.hook_cls(self.extension, [item1, item2])

        self.assertIn(item1, self.registry)
        self.assertIn(item2, self.registry)

    def test_initialize_handles_errors(self):
        """Testing BaseRegistryMultiItemHook.initialize unregisters all items
        on registration error
        """
        self.assertEqual(list(self.registry), [])

        item1 = DummyItem(123)
        item2 = DummyItem(123)

        with self.assertRaises(AlreadyRegisteredError):
            self.hook_cls(self.extension, [item1, item2])

        self.assertEqual(list(self.registry), [])

    def test_shutdown(self):
        """Testing BaseRegistryMultiItemHook.shutdown"""
        self.assertEqual(list(self.registry), [])

        item1 = DummyItem(123)
        item2 = DummyItem(456)
        self.hook_cls(self.extension, [item1, item2])

        self.extension.shutdown()
        self.assertEqual(list(self.registry), [])



"""Base class for a resource in an API."""

from __future__ import unicode_literals

import json
import logging
import warnings

from django.conf.urls import include, url
from django.core.exceptions import ObjectDoesNotExist
from django.core.urlresolvers import reverse
from django.db import models
from django.db.models.query import QuerySet
from django.http import (HttpResponseNotAllowed, HttpResponse,
                         HttpResponseNotModified)
from django.utils import six
from django.views.decorators.vary import vary_on_headers

from djblets.auth.ratelimit import (RATE_LIMIT_API_ANONYMOUS,
                                    RATE_LIMIT_API_AUTHENTICATED,
                                    get_usage_count)
from djblets.util.http import (build_not_modified_from_response,
                               encode_etag,
                               etag_if_none_match,
                               get_http_requested_mimetype,
                               get_modified_since,
                               set_etag,
                               set_last_modified)
from djblets.urls.patterns import never_cache_patterns
from djblets.webapi.auth.backends import check_login
from djblets.webapi.resources.registry import (get_resource_for_object,
                                               _class_to_resources,
                                               _name_to_resources)
from djblets.webapi.responses import (WebAPIResponse,
                                      WebAPIResponseError,
                                      WebAPIResponsePaginated)
from djblets.webapi.decorators import (SPECIAL_PARAMS,
                                       webapi_login_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST,
                                   LOGIN_FAILED,
                                   NOT_LOGGED_IN,
                                   PERMISSION_DENIED,
                                   RATE_LIMIT_EXCEEDED,
                                   WebAPIError)
from djblets.webapi.fields import IntFieldType

try:
    # Django >= 1.9
    from django.db.models.fields.related_descriptors import (
        ManyToManyDescriptor,
        ForwardManyToOneDescriptor)

    m2m_descriptors = (ManyToManyDescriptor,)
    fkey_descriptors = (ForwardManyToOneDescriptor,)
except ImportError:
    # Django < 1.9
    from django.db.models.fields.related import (
        ManyRelatedObjectsDescriptor,
        ReverseManyRelatedObjectsDescriptor,
        ReverseSingleRelatedObjectDescriptor)

    m2m_descriptors = (
        ManyRelatedObjectsDescriptor,
        ReverseManyRelatedObjectsDescriptor,
    )
    fkey_descriptors = (ReverseSingleRelatedObjectDescriptor,)


logger = logging.getLogger(__name__)


class WebAPIResource(object):
    """A resource handling HTTP operations for part of the API.

    A WebAPIResource is a RESTful resource living at a specific URL. It
    can represent either an object or a list of objects, and can respond
    to various HTTP methods (GET, POST, PUT, DELETE).

    Subclasses are expected to override functions and variables in order to
    provide specific functionality, such as modifying the resource or
    creating a new resource.

    For information on writing an API resource, see
    :ref:`writing-api-resources`.
    """

    # Configuration
    model = None
    fields = {}
    uri_object_key_regex = r'[0-9]+'
    uri_object_key = None
    model_object_key = 'pk'
    model_parent_key = None
    last_modified_field = None
    etag_field = None
    autogenerate_etags = False
    singleton = False
    list_child_resources = []
    item_child_resources = []
    allowed_methods = ('GET',)
    mimetype_vendor = None
    mimetype_list_resource_name = None
    mimetype_item_resource_name = None
    allowed_mimetypes = [
        {'list': mime, 'item': mime}
        for mime in WebAPIResponse.supported_mimetypes
    ]

    #: The class to use for paginated results in get_list.
    paginated_cls = WebAPIResponsePaginated

    # State
    method_mapping = {
        'GET': 'get',
        'POST': 'post',
        'PUT': 'put',
        'DELETE': 'delete',
    }

    _parent_resource = None
    _mimetypes_cache = None

    def __init__(self):
        _name_to_resources[self.name] = self
        _name_to_resources[self.name_plural] = self
        _class_to_resources[self.__class__] = self

        # Mark this class, and any subclasses, to be Web API handlers
        self.is_webapi_handler = True

        # Copy this list, because otherwise we may modify the class-level
        # version of it.
        self.allowed_mimetypes = list(self.allowed_mimetypes)

        if self.mimetype_vendor:
            # Generate list and item resource-specific mimetypes
            # for each supported mimetype, and add them as a pair to the
            # allowed mimetypes.
            for mimetype_pair in self.allowed_mimetypes:
                vend_mimetype_pair = {
                    'list': None,
                    'item': None,
                }

                for key, is_list in [('list', True), ('item', False)]:
                    if (key in mimetype_pair and
                        (mimetype_pair[key] in
                         WebAPIResponse.supported_mimetypes)):
                        vend_mimetype_pair[key] = \
                            self._build_resource_mimetype(mimetype_pair[key],
                                                          is_list)

                if vend_mimetype_pair['list'] or vend_mimetype_pair['item']:
                    self.allowed_mimetypes.append(vend_mimetype_pair)

    @vary_on_headers('Accept', 'Cookie')
    def __call__(self, request, api_format=None, *args, **kwargs):
        """Invokes the correct HTTP handler based on the type of request."""
        if not hasattr(request, '_djblets_webapi_object_cache'):
            request._djblets_webapi_object_cache = {}

        auth_result = check_login(request)

        if isinstance(auth_result, tuple):
            auth_success, auth_message, auth_headers = auth_result

            if not auth_success:
                err = LOGIN_FAILED

                if auth_message:
                    err = err.with_message(auth_message)

                return WebAPIResponseError(
                    request,
                    err=err,
                    headers=auth_headers or {},
                    api_format=api_format,
                    mimetype=self._build_error_mimetype(request))

        # Check to see if the user has been rate limited. Rate limits for
        # authentication failures have already been handled via the auth
        # backend, so this just checks for the API rate limits, if any.
        if auth_result:
            rate_limit_type = RATE_LIMIT_API_AUTHENTICATED
        else:
            rate_limit_type = RATE_LIMIT_API_ANONYMOUS

        usage = get_usage_count(request, increment=True,
                                limit_type=rate_limit_type)

        if (usage is not None and
            usage['count'] > usage['limit']):
            return WebAPIResponseError(
                request,
                err=RATE_LIMIT_EXCEEDED,
                headers={
                    'Retry-After': usage['time_left'],
                    'X-RateLimit-Limit': usage['limit'],
                },
                api_format=api_format,
                mimetype=self._build_error_mimetype(request))

        method = request.method

        if method == 'POST':
            # Not all clients can do anything other than GET or POST.
            # So, in the case of POST, we allow overriding the method
            # used.
            method = request.POST.get('_method', kwargs.get('_method', method))
        elif method == 'PUT':
            # Normalize the PUT data so we can get to it.
            # This is due to Django's treatment of PUT vs. POST. They claim
            # that PUT, unlike POST, is not necessarily represented as form
            # data, so they do not parse it. However, that gives us no clean
            # way of accessing the data. So we pretend it's POST for a second
            # in order to parse.
            #
            # This must be done only for legitimate PUT requests, not faked
            # ones using ?method=PUT.
            try:
                request.method = 'POST'
                request._load_post_and_files()
                request.method = 'PUT'
            except AttributeError:
                request.META['REQUEST_METHOD'] = 'POST'
                request._load_post_and_files()
                request.META['REQUEST_METHOD'] = 'PUT'

        request._djblets_webapi_method = method
        request._djblets_webapi_kwargs = kwargs
        request.PUT = request.POST

        if method in self.allowed_methods:
            if (method == "GET" and
                not self.singleton and
                (self.uri_object_key is None or
                 self.uri_object_key not in kwargs)):
                view = self.get_list
            else:
                view = getattr(self, self.method_mapping.get(method, None))
        else:
            view = None

        if view and six.callable(view):
            result = self.call_method_view(
                request, method, view, api_format=api_format, *args, **kwargs)

            if isinstance(result, WebAPIResponse):
                response = result
            elif isinstance(result, WebAPIError):
                response = WebAPIResponseError(
                    request,
                    err=result,
                    api_format=api_format,
                    mimetype=self._build_error_mimetype(request))
            elif isinstance(result, tuple):
                headers = {}

                if method == 'GET':
                    request_params = request.GET
                else:
                    request_params = request.POST

                if len(result) == 3:
                    headers = result[2]

                if 'Location' in headers:
                    extra_querystr = '&'.join([
                        '%s=%s' % (param, request_params[param])
                        for param in SPECIAL_PARAMS
                        if param in request_params
                    ])

                    if extra_querystr:
                        if '?' in headers['Location']:
                            headers['Location'] += '&' + extra_querystr
                        else:
                            headers['Location'] += '?' + extra_querystr

                if isinstance(result[0], WebAPIError):
                    response = WebAPIResponseError(
                        request,
                        err=result[0],
                        headers=headers,
                        extra_params=result[1],
                        api_format=api_format,
                        mimetype=self._build_error_mimetype(request))
                else:
                    response_args = self.build_response_args(request)
                    headers.update(response_args.pop('headers', {}))

                    response = WebAPIResponse(
                        request,
                        status=result[0],
                        obj=result[1],
                        headers=headers,
                        api_format=api_format,
                        encoder_kwargs=dict({
                            'calling_resource': self,
                        }, **kwargs),
                        **response_args)
            elif isinstance(result, HttpResponse):
                response = result
            else:
                raise AssertionError(result)

            if 200 <= response.status_code < 300:
                # Check if there's no explicit ETag in a GET/HEAD request and
                # the resource is told to generate missing ETags automatically.
                #
                # Starting in Djblets 2.0, we do this after a response payload
                # is generated instead of up-front, so that we don't serialize
                # the resource payload twice.
                if (self.autogenerate_etags and
                    method in ('GET', 'HEAD') and
                    'ETag' not in response):
                    # There's no ETag, so we'll build one based on the payload
                    # content and check to see if the client already has it.
                    etag = encode_etag(response.content)
                    set_etag(response, etag)

                    if self.are_cache_headers_current(request, etag=etag):
                        response = build_not_modified_from_response(response)
        else:
            response = HttpResponseNotAllowed(self.allowed_methods)

        return response

    @property
    def __name__(self):
        return self.__class__.__name__

    @property
    def name(self):
        """Returns the name of the object, used for keys in the payloads."""
        if not hasattr(self, '_name'):
            if self.model:
                self._name = self.model.__name__.lower()
            else:
                self._name = self.__name__.lower()

        return self._name

    @property
    def name_plural(self):
        """Returns the plural name of the object, used for lists."""
        if not hasattr(self, '_name_plural'):
            if self.singleton:
                self._name_plural = self.name
            else:
                self._name_plural = self.name + 's'

        return self._name_plural

    @property
    def item_result_key(self):
        """Returns the key for single objects in the payload."""
        return self.name

    @property
    def list_result_key(self):
        """Returns the key for lists of objects in the payload."""
        return self.name_plural

    @property
    def uri_name(self):
        """Returns the name of the resource in the URI.

        This can be overridden when the name in the URI needs to differ
        from the name used for the resource.
        """
        return self.name_plural.replace('_', '-')

    @property
    def link_name(self):
        """Returns the name of the resource for use in a link.

        This can be overridden when the name in the link needs to differ
        from the name used for the resource.
        """
        return self.name_plural

    def call_method_view(self, request, method, view, *args, **kwargs):
        """Calls the given method view.

        This will just call the given view by default, passing in all
        args and kwargs.

        This can be overridden by subclasses to perform additional
        checks or pass additional data to the view.
        """
        return view(request, *args, **kwargs)

    def build_response_args(self, request):
        is_list = (request._djblets_webapi_method == 'GET' and
                   not self.singleton and
                   (self.uri_object_key is None or
                    self.uri_object_key not in request._djblets_webapi_kwargs))

        if is_list:
            key = 'list'
        else:
            key = 'item'

        supported_mimetypes = [
            mime[key]
            for mime in self.allowed_mimetypes
            if mime.get(key)
        ]

        mimetype = get_http_requested_mimetype(request, supported_mimetypes)

        if (self.mimetype_vendor and
            mimetype in WebAPIResponse.supported_mimetypes):
            mimetype = self._build_resource_mimetype(mimetype, is_list)

        response_args = {
            'supported_mimetypes': supported_mimetypes,
            'mimetype': mimetype,
        }

        if is_list:
            for mimetype_pair in self.allowed_mimetypes:
                if (mimetype_pair.get('list') == mimetype and
                    mimetype_pair.get('item')):
                    response_args['headers'] = {
                        'Item-Content-Type': mimetype_pair['item'],
                    }
                    break

        return response_args

    def get_object(self, request, id_field=None, *args, **kwargs):
        """Returns an object, given captured parameters from a URL.

        This will perform a query for the object, taking into account
        ``model_object_key``, ``uri_object_key``, and any captured parameters
        from the URL.

        This requires that ``model`` and ``uri_object_key`` be set.

        Throws django.core.exceptions.ObjectDoesNotExist if the requested
        object does not exist.
        """
        assert self.model
        assert self.singleton or self.uri_object_key

        if self.singleton:
            cache_key = '%d' % id(self)
        else:
            id_field = id_field or self.model_object_key
            object_id = kwargs[self.uri_object_key]
            cache_key = '%d:%s:%s' % (id(self), id_field, object_id)

        if cache_key in request._djblets_webapi_object_cache:
            return request._djblets_webapi_object_cache[cache_key]

        if 'is_list' in kwargs:
            # Don't pass this in to _get_queryset, since we're not fetching
            # a list, and don't want the extra optimizations for lists to
            # kick in.
            del kwargs['is_list']

        queryset = self._get_queryset(request, *args, **kwargs)

        if self.singleton:
            obj = queryset.get()
        else:
            obj = queryset.get(**{
                id_field: object_id,
            })

        request._djblets_webapi_object_cache[cache_key] = obj

        return obj

    def post(self, *args, **kwargs):
        """Handles HTTP POSTs.

        This is not meant to be overridden unless there are specific needs.

        This will invoke ``create`` if doing an HTTP POST on a list resource.

        By default, an HTTP POST is not allowed on individual object
        resourcces.
        """

        if 'POST' not in self.allowed_methods:
            return HttpResponseNotAllowed(self.allowed_methods)

        if (self.uri_object_key is None or
            kwargs.get(self.uri_object_key, None) is None):
            return self.create(*args, **kwargs)

        # Don't allow POSTs on children by default.
        allowed_methods = list(self.allowed_methods)
        allowed_methods.remove('POST')

        return HttpResponseNotAllowed(allowed_methods)

    def put(self, request, *args, **kwargs):
        """Handles HTTP PUTs.

        This is not meant to be overridden unless there are specific needs.

        This will just invoke ``update``.
        """
        return self.update(request, *args, **kwargs)

    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED)
    def get(self, request, api_format, *args, **kwargs):
        """Handles HTTP GETs to individual object resources.

        By default, this will check for access permissions and query for
        the object. It will then return a serialized form of the object.

        This may need to be overridden if needing more complex logic.
        """
        if (not self.model or
            (self.uri_object_key is None and not self.singleton)):
            return HttpResponseNotAllowed(self.allowed_methods)

        try:
            obj = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_access_permissions(request, obj, *args, **kwargs):
            return self.get_no_access_error(request, obj=obj, *args, **kwargs)

        last_modified_timestamp = self.get_last_modified(request, obj)
        etag = self.get_etag(request, obj, **kwargs)

        if self.are_cache_headers_current(request, last_modified_timestamp,
                                          etag):
            response = HttpResponseNotModified()
        else:
            data = {
                self.item_result_key: self.serialize_object(obj,
                                                            request=request,
                                                            *args, **kwargs),
            }

            response = WebAPIResponse(request,
                                      status=200,
                                      obj=data,
                                      api_format=api_format,
                                      **self.build_response_args(request))

        if last_modified_timestamp:
            set_last_modified(response, last_modified_timestamp)

        if etag:
            set_etag(response, etag)

        return response

    @webapi_response_errors(NOT_LOGGED_IN, PERMISSION_DENIED, DOES_NOT_EXIST)
    @webapi_request_fields(
        optional={
            'start': {
                'type': IntFieldType,
                'description': 'The 0-based index of the first result in '
                               'the list. The start index is usually the '
                               'previous start index plus the number of '
                               'previous results. By default, this is 0.',
            },
            'max-results': {
                'type': IntFieldType,
                'description': 'The maximum number of results to return in '
                               'this list. By default, this is 25. There is '
                               'a hard limit of 200; if you need more than '
                               '200 results, you will need to make more '
                               'than one request, using the "next" '
                               'pagination link.',
            }
        },
        allow_unknown=True
    )
    def get_list(self, request, *args, **kwargs):
        """Handles HTTP GETs to list resources.

        By default, this will query for a list of objects and return the
        list in a serialized form.
        """
        data = {
            'links': self.get_links(self.list_child_resources,
                                    request=request, *args, **kwargs),
        }

        if not self.has_list_access_permissions(request, *args, **kwargs):
            return self.get_no_access_error(request, *args, **kwargs)

        if self.model:
            try:
                queryset = self._get_queryset(request, is_list=True,
                                              *args, **kwargs)
            except ObjectDoesNotExist:
                return DOES_NOT_EXIST

            return self.paginated_cls(
                request,
                queryset=queryset,
                results_key=self.list_result_key,
                serialize_object_func=lambda obj:
                    self.get_serializer_for_object(obj).serialize_object(
                        obj, request=request, *args, **kwargs),
                extra_data=data,
                **self.build_response_args(request))
        else:
            return 200, data

    @webapi_login_required
    def create(self, request, api_format, *args, **kwargs):
        """Handles HTTP POST requests to list resources.

        This is used to create a new object on the list, given the
        data provided in the request. It should usually return
        HTTP 201 Created upon success.

        By default, this returns HTTP 405 Method Not Allowed.
        """
        return HttpResponseNotAllowed(self.allowed_methods)

    @webapi_login_required
    def update(self, request, api_format, *args, **kwargs):
        """Handles HTTP PUT requests to object resources.

        This is used to update an object, given full or partial data provided
        in the request. It should usually return HTTP 200 OK upon success.

        By default, this returns HTTP 405 Method Not Allowed.
        """
        return HttpResponseNotAllowed(self.allowed_methods)

    @webapi_login_required
    @webapi_response_errors(DOES_NOT_EXIST, NOT_LOGGED_IN, PERMISSION_DENIED)
    def delete(self, request, api_format, *args, **kwargs):
        """Handles HTTP DELETE requests to object resources.

        This is used to delete an object, if the user has permissions to
        do so.

        By default, this deletes the object and returns HTTP 204 No Content.
        """
        if not self.model or self.uri_object_key is None:
            return HttpResponseNotAllowed(self.allowed_methods)

        try:
            obj = self.get_object(request, *args, **kwargs)
        except ObjectDoesNotExist:
            return DOES_NOT_EXIST

        if not self.has_delete_permissions(request, obj, *args, **kwargs):
            return self.get_no_access_error(request, obj=obj, *args, **kwargs)

        obj.delete()

        return 204, {}

    def get_queryset(self, request, is_list=False, *args, **kwargs):
        """Returns a queryset used for querying objects or lists of objects.

        Throws django.core.exceptions.ObjectDoesNotExist if the requested
        object does not exist.

        This can be overridden to filter the object list, such as for hiding
        non-public objects.

        The ``is_list`` parameter can be used to specialize the query based
        on whether an individual object or a list of objects is being queried.
        """
        return self.model.objects.all()

    def get_url_patterns(self):
        """Returns the Django URL patterns for this object and its children.

        This is used to automatically build up the URL hierarchy for all
        objects. Projects should call this for top-level resources and
        return them in the ``urls.py`` files.
        """
        urlpatterns = never_cache_patterns(
            url(r'^$', self, name=self._build_named_url(self.name_plural)),
        )

        for resource in self.list_child_resources:
            resource._parent_resource = self
            child_regex = r'^' + resource.uri_name + r'/'
            urlpatterns += [
                url(child_regex, include(resource.get_url_patterns())),
            ]

        if self.uri_object_key or self.singleton:
            # If the resource has particular items in it...
            if self.uri_object_key:
                base_regex = r'^(?P<%s>%s)/' % (self.uri_object_key,
                                                self.uri_object_key_regex)
            elif self.singleton:
                base_regex = r'^'

            urlpatterns += never_cache_patterns(
                url(base_regex + r'$', self,
                    name=self._build_named_url(self.name))
            )

            for resource in self.item_child_resources:
                resource._parent_resource = self
                child_regex = base_regex + resource.uri_name + r'/'
                urlpatterns += [
                    url(child_regex, include(resource.get_url_patterns())),
                ]

        return urlpatterns

    def has_access_permissions(self, request, obj, *args, **kwargs):
        """Returns whether or not the user has read access to this object."""
        return True

    def has_list_access_permissions(self, request, *args, **kwargs):
        """Returns whether or not the user has read access to this list."""
        if self._parent_resource and self.model_parent_key:
            try:
                parent_obj = self._parent_resource.get_object(
                    request, *args, **kwargs)

                return self._parent_resource.has_access_permissions(
                    request, parent_obj, *args, **kwargs)
            except:
                # Other errors, like Does Not Exist, should be caught
                # separately. As of here, we'll allow it to pass, so that
                # the error isn't a Permission Denied when it should be
                # a Does Not Exist.
                pass

        return True

    def has_modify_permissions(self, request, obj, *args, **kwargs):
        """Returns whether or not the user can modify this object."""
        return False

    def has_delete_permissions(self, request, obj, *args, **kwargs):
        """Returns whether or not the user can delete this object."""
        return False

    def get_link_serializer(self, field):
        """Return the function to use for serializing a link field."""
        serialize_link_func = getattr(self, 'serialize_%s_link' % field,
                                      None)

        if not serialize_link_func or not six.callable(serialize_link_func):
            serialize_link_func = self.serialize_link

        return serialize_link_func

    def serialize_link(self, obj, *args, **kwargs):
        """Serialize a link to the object into a Python dictionary."""
        resource = self.get_serializer_for_object(obj)
        assert resource

        return {
            'method': 'GET',
            'href': resource.get_href(obj, *args, **kwargs),
            'title': resource.get_object_title(obj, *args, **kwargs),
        }

    def get_object_title(self, obj, request=None, *args, **kwargs):
        """Return the object's title.

        By default, this returns the object's unicode representation.

        Args:
            obj (object):
                The object to serialize.

            request (django.http.HttpRequest):
                The current request.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            unicode:
            The object's title.
        """
        return six.text_type(obj)

    def serialize_object(self, obj, *args, **kwargs):
        """Serializes the object into a Python dictionary."""
        request = kwargs.get('request', None)

        if request:
            if not hasattr(request, '_djblets_webapi_serialize_cache'):
                request._djblets_webapi_serialize_cache = {}

            if obj in request._djblets_webapi_serialize_cache:
                return self._clone_serialized_object(
                    request._djblets_webapi_serialize_cache[obj])

        only_fields = self.get_only_fields(request)
        only_links = self.get_only_links(request)

        data = {}
        links = {}
        expand_info = {}

        if only_links != []:
            links = self.get_links(self.item_child_resources, obj,
                                   *args, **kwargs)

        if hasattr(request, '_djblets_webapi_expanded_resources'):
            expanded_resources = request._djblets_webapi_expanded_resources
        else:
            expand = request.GET.get('expand', request.POST.get('expand', ''))
            expanded_resources = expand.split(',')
            request._djblets_webapi_expanded_resources = expanded_resources

        if expanded_resources:
            requested_mimetype = get_http_requested_mimetype(
                request, WebAPIResponse.supported_mimetypes)
        else:
            requested_mimetype = None

        # Make a copy of the list of expanded resources. We'll be temporarily
        # removing items as we recurse down into any nested objects, to
        # prevent infinite loops. We'll want to make sure we don't
        # permanently remove these entries, or subsequent list items will
        # be affected.
        orig_expanded_resources = list(expanded_resources)

        for field in six.iterkeys(self.fields):
            can_include_field = only_fields is None or field in only_fields
            expand_field = field in expanded_resources

            # If we're limiting fields and this one isn't explicitly included,
            # then we're only going to want to process it if there's a chance
            # it'll be linked (as opposed to being expanded).
            if not can_include_field and expand_field:
                continue

            serialize_func = getattr(self, "serialize_%s_field" % field, None)

            if serialize_func and six.callable(serialize_func):
                value = serialize_func(obj, request=request)
            else:
                value = getattr(obj, field)

                if isinstance(value, models.Manager):
                    if not can_include_field:
                        # This field isn't a single Model, so it can't be
                        # linked below. We can safely bail now before talking
                        # to the database.
                        continue

                    value = value.all()
                elif isinstance(value, models.ForeignKey):
                    value = value.get()

            # Make sure that any given field expansion only applies once. This
            # prevents infinite recursion in the case where there's a loop in
            # the object graph.
            #
            # We'll be restoring these values once we're done serializing
            # objects.
            if expand_field:
                request._djblets_webapi_expanded_resources.remove(field)

            if isinstance(value, models.Model) and not expand_field:
                serialize_link_func = self.get_link_serializer(field)

                links[field] = serialize_link_func(value, *args, **kwargs)
            elif can_include_field:
                if isinstance(value, QuerySet) and not expand_field:
                    serialize_link_func = self.get_link_serializer(field)

                    data[field] = [
                        serialize_link_func(o, *args, **kwargs)
                        for o in value
                    ]
                elif isinstance(value, QuerySet):
                    objects = list(value)

                    if objects:
                        resource = self.get_serializer_for_object(objects[0])

                        expand_info[field] = {
                            'item_mimetype': resource._build_resource_mimetype(
                                mimetype=requested_mimetype,
                                is_list=False),
                        }

                        data[field] = [
                            resource.serialize_object(o, *args, **kwargs)
                            for o in objects
                        ]
                    else:
                        data[field] = []
                elif isinstance(value, models.Model):
                    resource = self.get_serializer_for_object(value)

                    expand_info[field] = {
                        'item_mimetype': resource._build_resource_mimetype(
                            mimetype=requested_mimetype,
                            is_list=False),
                    }

                    data[field] = resource.serialize_object(
                        value, *args, **kwargs)
                else:
                    data[field] = value

        for resource_name in expanded_resources:
            if (resource_name not in links or
                (only_fields is not None and
                 resource_name not in only_fields)):
                continue

            # Try to find the resource from the child list.
            found = False

            for resource in self.item_child_resources:
                if resource_name in [resource.name, resource.name_plural]:
                    found = True
                    break

            if not found or not resource.model:
                continue

            extra_kwargs = {
                self.uri_object_key: getattr(obj, self.model_object_key),
            }
            extra_kwargs.update(**kwargs)
            extra_kwargs.update(self.get_href_parent_ids(obj, **kwargs))

            data[resource_name] = [
                resource.serialize_object(o, *args, **kwargs)
                for o in resource._get_queryset(
                    is_list=True, *args, **extra_kwargs)
            ]

            expand_info[resource_name] = {
                'list_url': links[resource_name]['href'],
                'list_mimetype': resource._build_resource_mimetype(
                    mimetype=requested_mimetype,
                    is_list=True),
                'item_mimetype': resource._build_resource_mimetype(
                    mimetype=requested_mimetype,
                    is_list=False),
            }

            del links[resource_name]

        if only_links is None:
            data['links'] = links
        elif only_links != []:
            data['links'] = dict([
                (link_name, link_info)
                for link_name, link_info in six.iteritems(links)
                if link_name in only_links
            ])

        if expand_info:
            data['_expanded'] = expand_info

        # Now that we're done serializing, restore the list of expanded
        # resource for the next call.
        request._djblets_webapi_expanded_resources = orig_expanded_resources

        if request:
            request._djblets_webapi_serialize_cache[obj] = \
                self._clone_serialized_object(data)

        return data

    def get_only_fields(self, request):
        """Returns the list of the only fields that the payload should include.

        If the user has requested that no fields should be provided, this
        will return an empty list.

        If all fields will be included in the payload, this will return None.
        """
        return self._get_only_items(request, 'only-fields', 'only_fields')

    def get_only_links(self, request):
        """Returns the list of the only links that the payload should include.

        If the user has requested that no links should be provided, this
        will return an empty list.

        If all links will be included in the payload, this will return None.
        """
        return self._get_only_items(request, 'only-links', 'only_links')

    def get_serializer_for_object(self, obj):
        """Returns the serializer used to serialize an object.

        This is called when serializing objects for payloads returned
        by this resource instance. It must return the resource instance
        that will be responsible for serializing the given object for the
        payload.

        By default, this calls ``get_resource_for_object`` to find the
        appropriate resource.
        """
        return get_resource_for_object(obj)

    def get_links(self, resources=[], obj=None, request=None,
                  *args, **kwargs):
        """Returns a dictionary of links coming off this resource.

        The resulting links will point to the resources passed in
        ``resources``, and will also provide special resources for
        ``self`` (which points back to the official location for this
        resource) and one per HTTP method/operation allowed on this
        resource.
        """
        links = {}
        base_href = None

        if obj:
            base_href = self.get_href(obj, request, *args, **kwargs)

        if not base_href:
            # We may have received None from the URL above.
            if request:
                base_href = request.build_absolute_uri()
            else:
                base_href = ''

        links['self'] = {
            'method': 'GET',
            'href': base_href,
        }

        # base_href without any query arguments.
        i = base_href.find('?')

        if i != -1:
            clean_base_href = base_href[:i]
        else:
            clean_base_href = base_href

        if 'POST' in self.allowed_methods and not obj:
            links['create'] = {
                'method': 'POST',
                'href': clean_base_href,
            }

        if 'PUT' in self.allowed_methods and obj:
            links['update'] = {
                'method': 'PUT',
                'href': clean_base_href,
            }

        if 'DELETE' in self.allowed_methods and obj:
            links['delete'] = {
                'method': 'DELETE',
                'href': clean_base_href,
            }

        for resource in resources:
            links[resource.link_name] = {
                'method': 'GET',
                'href': '%s%s/' % (clean_base_href, resource.uri_name),
            }

        for key, info in six.iteritems(
                self.get_related_links(obj, request, *args, **kwargs)):
            links[key] = {
                'method': info['method'],
                'href': info['href'],
            }

            if 'title' in info:
                links[key]['title'] = info['title']

        return links

    def get_related_links(self, obj=None, request=None, *args, **kwargs):
        """Returns links related to this resource.

        The result should be a dictionary of link names to a dictionary of
        information. The information should contain:

        * 'method' - The HTTP method
        * 'href' - The URL
        * 'title' - The title of the link (optional)
        * 'resource' - The WebAPIResource instance
        * 'list-resource' - True if this links to a list resource (optional)
        """
        return {}

    def get_href(self, obj, request, *args, **kwargs):
        """Returns the URL for this object."""
        if not self.uri_object_key:
            return None

        href_kwargs = {
            self.uri_object_key: getattr(obj, self.model_object_key),
        }
        href_kwargs.update(self.get_href_parent_ids(obj, **kwargs))

        return self.get_item_url(request=request, **href_kwargs)

    def get_list_url(self, **kwargs):
        """Return the URL to the list version of this resource.

        This will generate a URL for the list resource, given the provided
        arguments for the URL pattern.

        Args:
            kwargs (dict): The keyword arguments needed for URL resolution.

        Returns:
            unicode: The resulting absolute URL to the list resource.
        """
        return self.build_resource_url(self.name_plural, **kwargs)

    def get_item_url(self, **kwargs):
        """Return the URL to the item version of this resource.

        This will generate a URL for the item resource, given the provided
        arguments for the URL pattern.

        Args:
            kwargs (dict): The keyword arguments needed for URL resolution.

        Returns:
            unicode: The resulting absolute URL to the item resource.
        """
        return self.build_resource_url(self.name, **kwargs)

    def build_resource_url(self, name, request=None, **kwargs):
        """Build a resource URL for the given name and keyword arguments.

        This can be overridden by subclasses that have special requirements
        for URL resolution.

        Args:
            name (unicode):
                The name of the resource.

            request (HttpRequest):
                The HTTP request from the client.

            kwargs (dict):
                The keyword arguments needed for URL resolution.

        Returns:
            unicode: The resulting absolute URL to the resource.
        """
        url = reverse(self._build_named_url(name), kwargs=kwargs)

        if request:
            url = request.build_absolute_uri(url)

        return url

    def get_href_parent_ids(self, obj, **kwargs):
        """Returns a dictionary mapping parent object keys to their values for
        an object.
        """
        parent_ids = {}

        if self._parent_resource and self.model_parent_key:
            parent_obj = self.get_parent_object(obj)
            parent_ids = self._parent_resource.get_href_parent_ids(
                parent_obj, **kwargs)

            if self._parent_resource.uri_object_key:
                parent_ids[self._parent_resource.uri_object_key] = \
                    getattr(parent_obj, self._parent_resource.model_object_key)

        return parent_ids

    def get_parent_object(self, obj):
        """Returns the parent of an object.

        By default, this uses ``model_parent_key`` to figure out the parent,
        but it can be overridden for more complex behavior.
        """
        parent_obj = getattr(obj, self.model_parent_key)

        if isinstance(parent_obj, (models.Manager, models.ForeignKey)):
            parent_obj = parent_obj.get()

        return parent_obj

    def get_last_modified(self, request, obj):
        """Returns the last modified timestamp of an object.

        By default, this uses ``last_modified_field`` to determine what
        field in the model represents the last modified timestamp of
        the object.

        This can be overridden for more complex behavior.
        """
        if self.last_modified_field:
            return getattr(obj, self.last_modified_field)

        return None

    def get_etag(self, request, obj, *args, **kwargs):
        """Returns the ETag representing the state of the object.

        By default, this uses ``etag_field`` to determine what field in
        the model is unique enough to represent the state of the object.

        This can be overridden for more complex behavior. Any overridden
        functions should make sure to pass the result through
        ``encode_etag`` before returning a value.
        """
        if self.etag_field:
            return self.encode_etag(
                request,
                six.text_type(getattr(obj, self.etag_field)))

        return None

    def encode_etag(self, request, etag, *args, **kwargs):
        """Encodes an ETag for usage in a header.

        This will take a precomputed ETag, augment it with additional
        information, encode it as a SHA1, and return it.
        """
        return encode_etag('%s:%s' % (request.user.username, etag))

    def generate_etag(self, obj, fields, request, encode_etag=True, **kwargs):
        """Generate an ETag from the serialized values of all given fields.

        .. deprecated:: 2.0

           This method is no longer called internally, as the auto-generation
           logic for ETags has changed. It's left here for any resources that
           may still call it, and is scheduled to be removed in a future
           release.

        Args:
            obj (object):
                The object being serialized.

            fields (list of unicode, unused):
                An unused list of field names. You can provide any value here.
                It will be ignored.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            encode_etag (bool, optional):
                Whether to encode the ETag to a SHA.

            **kwargs (dict):
                Extra keyword arguments for :py:meth:`serialize_object`.

        Returns:
            unicode:
            The generated ETag.
        """
        warnings.warn(
            'WebAPIResource.generate_etag is deprecated and no longer called '
            'by default. Please provide your own ETag generation in '
            'get_etag().',
            DeprecationWarning)

        etag = repr(self.serialize_object(obj, request=request, **kwargs))

        if encode_etag:
            etag = self.encode_etag(request, etag)

        return etag

    def are_cache_headers_current(self, request, last_modified=None,
                                  etag=None):
        """Determines if cache headers from the client are current.

        This will compare the optionally-provided timestamp and ETag against
        any conditional cache headers sent by the client to determine if
        the headers are current. If they are, the caller can return
        HttpResponseNotModified instead of a payload.
        """
        return ((last_modified and
                 get_modified_since(request, last_modified)) or
                (etag and etag_if_none_match(request, etag)))

    def get_no_access_error(self, request, *args, **kwargs):
        """Returns an appropriate error when access is denied.

        By default, this will return PERMISSION_DENIED if the user is logged
        in, and NOT_LOGGED_IN if the user is anonymous.

        Subclasses can override this to return different or more detailed
        errors.
        """
        if request.user.is_authenticated():
            logger.warning('%s %s: user %s does not have '
                           'permission to access this resource.',
                           request.method, request.path,
                           request.user.username)
            return PERMISSION_DENIED
        else:
            return NOT_LOGGED_IN

    def _build_resource_mimetype(self, mimetype, is_list):
        if is_list:
            resource_name = (self.mimetype_list_resource_name or
                             self.name_plural.replace('_', '-'))
        else:
            resource_name = (self.mimetype_item_resource_name or
                             self.name.replace('_', '-'))

        return self._build_vendor_mimetype(mimetype, resource_name)

    def _build_error_mimetype(self, request):
        mimetype = get_http_requested_mimetype(
            request, WebAPIResponse.supported_mimetypes)

        if self.mimetype_vendor:
            mimetype = self._build_vendor_mimetype(mimetype, 'error')

        return mimetype

    def _build_vendor_mimetype(self, mimetype, name):
        parts = mimetype.split('/')

        return '%s/vnd.%s.%s+%s' % (parts[0],
                                    self.mimetype_vendor,
                                    name,
                                    parts[1])

    def _build_named_url(self, name):
        """Builds a Django URL name from the provided name."""
        return '%s-resource' % name.replace('_', '-')

    def _get_only_items(self, request, query_param_name, post_field_name):
        if request:
            only = request.GET.get(query_param_name,
                                   request.POST.get(post_field_name, None))

            if only is not None:
                if only:
                    return only.split(',')
                else:
                    return []

        return None

    def _get_queryset(self, request, is_list=False, *args, **kwargs):
        """Returns an optimized queryset.

        This calls out to the resource's get_queryset(), and then performs
        some optimizations to better fetch related objects, reducing future
        lookups in this request.
        """
        queryset = self.get_queryset(request, is_list=is_list, *args, **kwargs)

        if not hasattr(self, '_select_related_fields'):
            self._select_related_fields = []

            for field in six.iterkeys(self.fields):
                if hasattr(self, 'serialize_%s_field' % field):
                    continue

                field_type = getattr(self.model, field, None)

                if field_type and isinstance(field_type, fkey_descriptors):
                    self._select_related_fields.append(field)

        if self._select_related_fields:
            queryset = \
                queryset.select_related(*self._select_related_fields)

        if is_list:
            if not hasattr(self, '_prefetch_related_fields'):
                self._prefetch_related_fields = []

                for field in six.iterkeys(self.fields):
                    if hasattr(self, 'serialize_%s_field' % field):
                        continue

                    field_type = getattr(self.model, field, None)

                    if field_type and isinstance(field_type, m2m_descriptors):
                        self._prefetch_related_fields.append(field)

            if self._prefetch_related_fields:
                queryset = \
                    queryset.prefetch_related(*self._prefetch_related_fields)

        return queryset

    def _clone_serialized_object(self, obj):
        """Clone a serialized object, for storing in the cache.

        This works similarly to deepcopy(), but is smart enough to only
        copy primitive types (dictionaries, lists, etc.) and won't
        interfere with model instances.

        deepcopy() should be smart enough to do that, and is documented
        as being smart enough, but Django models provide some functions
        that cause deepcopy() to dig in further than it should, eventually
        breaking in some cases.

        If you want the job done right, do it yourself.
        """
        if isinstance(obj, dict):
            return dict(
                (key, self._clone_serialized_object(value))
                for key, value in six.iteritems(obj)
            )
        elif isinstance(obj, list):
            return [
                self._clone_serialized_object(value)
                for value in obj
            ]
        else:
            return obj



"""Unit tests for djblets.extensions.extension.ExtensionInfo."""

from __future__ import unicode_literals

import os

from django.conf import settings

from djblets.extensions.extension import Extension, ExtensionInfo
from djblets.extensions.tests.base import FakeEntryPoint
from djblets.testing.testcases import TestCase


class ExtensionInfoTests(TestCase):
    """Unit tests for djblets.extensions.extension.ExtensionInfo."""

    def test_create_from_entrypoint(self):
        """Testing ExtensionInfo.create_from_entrypoint"""
        module_name = 'test_extension.dummy.submodule'
        package_name = 'DummyExtension'
        extension_id = '%s:DummyExtension' % module_name

        class TestExtension(Extension):
            __module__ = module_name
            id = extension_id

        entrypoint = FakeEntryPoint(TestExtension, project_name=package_name)
        extension_info = ExtensionInfo.create_from_entrypoint(entrypoint,
                                                              TestExtension)

        self._check_extension_info(extension_info=extension_info,
                                   app_name='test_extension.dummy',
                                   package_name=package_name,
                                   extension_id=extension_id,
                                   metadata=entrypoint.dist.metadata)

    def test_create_from_entrypoint_with_custom_metadata(self):
        """Testing ExtensionInfo.create_from_entrypoint with custom
        Extension.metadata
        """
        package_name = 'DummyExtension'
        module_name = 'test_extension.dummy.submodule'
        extension_id = '%s:DummyExtension' % module_name

        class TestExtension(Extension):
            __module__ = module_name
            id = extension_id
            metadata = {
                'Name': 'OverrideName',
                'Version': '3.14159',
                'Summary': 'Lorem ipsum dolor sit amet.',
                'Description': 'Tempus fugit.',
                'License': 'None',
                'Home-page': 'http://127.0.0.1/',
            }

        entrypoint = FakeEntryPoint(TestExtension, project_name=package_name)
        extension_info = ExtensionInfo.create_from_entrypoint(entrypoint,
                                                              TestExtension)

        expected_metadata = entrypoint.dist.metadata.copy()
        expected_metadata.update(TestExtension.metadata)

        self._check_extension_info(extension_info=extension_info,
                                   app_name='test_extension.dummy',
                                   package_name=package_name,
                                   extension_id=extension_id,
                                   metadata=expected_metadata)

    def _check_extension_info(self, extension_info, app_name, package_name,
                              extension_id, metadata):
        htdocs_path = os.path.join(settings.MEDIA_ROOT, 'ext', package_name)
        static_path = os.path.join(settings.STATIC_ROOT, 'ext', extension_id)

        self.assertEqual(extension_info.app_name, app_name)
        self.assertEqual(extension_info.author, metadata['Author'])
        self.assertEqual(extension_info.author_email, metadata['Author-email'])
        self.assertEqual(extension_info.description, metadata['Description'])
        self.assertFalse(extension_info.enabled)
        self.assertEqual(extension_info.installed_htdocs_path, htdocs_path)
        self.assertEqual(extension_info.installed_static_path, static_path)
        self.assertFalse(extension_info.installed)
        self.assertEqual(extension_info.license, metadata['License'])
        self.assertEqual(extension_info.metadata, metadata)
        self.assertEqual(extension_info.name, metadata['Name'])
        self.assertEqual(extension_info.package_name, package_name)
        self.assertEqual(extension_info.summary, metadata['Summary'])
        self.assertEqual(extension_info.url, metadata['Home-page'])
        self.assertEqual(extension_info.version, metadata['Version'])



from __future__ import unicode_literals

import json
from xml.sax.saxutils import XMLGenerator

from django.conf import settings
from django.contrib.auth.models import User, Group
from django.db.models.query import QuerySet
from django.utils import six
from django.utils.six.moves import cStringIO as StringIO

from djblets.util.serializers import DjbletsJSONEncoder


class WebAPIEncoder(object):
    """Encodes an object into a dictionary of fields and values.

    This object is used for both JSON and XML API formats.

    Projects can subclass this to provide representations of their objects.
    To make use of a encoder, add the path to the encoder class to
    the project's ``settings.WEB_API_ENCODERS`` list.

    For example:

    .. code-block:: python

        WEB_API_ENCODERS = (
            'myproject.webapi.MyEncoder',
        )
    """

    def encode(self, o, *args, **kwargs):
        """Encodes an object.

        This is expected to return either a dictionary or a list. If the
        object being encoded is not supported, return None, or call
        the superclass's encode method.
        """
        return None


class BasicAPIEncoder(WebAPIEncoder):
    """A basic encoder that encodes standard types.

    This supports encoding of dates, times, QuerySets, Users, and Groups.
    """
    def encode(self, o, *args, **kwargs):
        if isinstance(o, QuerySet):
            return list(o)
        elif isinstance(o, User):
            return {
                'id': o.id,
                'username': o.username,
                'first_name': o.first_name,
                'last_name': o.last_name,
                'fullname': o.get_full_name(),
                'email': o.email,
                'url': o.get_absolute_url(),
            }
        elif isinstance(o, Group):
            return {
                'id': o.id,
                'name': o.name,
            }
        else:
            try:
                return DjbletsJSONEncoder().default(o)
            except TypeError:
                return None


class ResourceAPIEncoder(WebAPIEncoder):
    """An encoder that encodes objects based on registered resources."""
    def encode(self, o, *args, **kwargs):
        if isinstance(o, QuerySet):
            return list(o)
        else:
            calling_resource = kwargs.pop('calling_resource', None)

            if calling_resource:
                serializer = calling_resource.get_serializer_for_object(o)
            else:
                from djblets.webapi.resources import get_resource_for_object

                serializer = get_resource_for_object(o)

            if serializer:
                return serializer.serialize_object(o, *args, **kwargs)
            else:
                try:
                    return DjbletsJSONEncoder().default(o)
                except TypeError:
                    return None


class JSONEncoderAdapter(json.JSONEncoder):
    """Adapts a WebAPIEncoder to be used with json.

    This takes an existing encoder and makes it available to use as a
    json.JSONEncoder. This is used internally when generating JSON from a
    WebAPIEncoder, but can be used in other projects for more specific
    purposes as well.
    """
    def __init__(self, encoder, *args, **kwargs):
        super(JSONEncoderAdapter, self).__init__(
            sort_keys=kwargs.pop('sort_keys', True),
            *args, **kwargs)
        self.encoder = encoder

    def encode(self, o, *args, **kwargs):
        self.encode_args = args
        self.encode_kwargs = kwargs
        return super(JSONEncoderAdapter, self).encode(o)

    def default(self, o):
        """Encodes an object using the supplied WebAPIEncoder.

        If the encoder is unable to encode this object, a TypeError is raised.
        """
        result = self.encoder.encode(o, *self.encode_args,
                                     **self.encode_kwargs)

        if result is None:
            raise TypeError("%r is not JSON serializable" % (o,))

        return result


class XMLEncoderAdapter(object):
    """Adapts a WebAPIEncoder to output XML.

    This takes an existing encoder and adapts it to output a simple XML format.
    """
    def __init__(self, encoder, *args, **kwargs):
        self.encoder = encoder

    def encode(self, o, *args, **kwargs):
        self.level = 0
        self.doIndent = False

        stream = StringIO()
        self.xml = XMLGenerator(stream, settings.DEFAULT_CHARSET)
        self.xml.startDocument()
        self.startElement("rsp")
        self.__encode(o, *args, **kwargs)
        self.endElement("rsp")
        self.xml.endDocument()
        self.xml = None

        return stream.getvalue()

    def __encode(self, o, *args, **kwargs):
        if isinstance(o, dict):
            for key, value in six.iteritems(o):
                attrs = {}

                if isinstance(key, six.integer_types):
                    attrs['value'] = str(key)
                    key = 'int'

                self.startElement(key, attrs)
                self.__encode(value, *args, **kwargs)
                self.endElement(key)
        elif isinstance(o, (tuple, list)):
            self.startElement("array")

            for i in o:
                self.startElement("item")
                self.__encode(i, *args, **kwargs)
                self.endElement("item")

            self.endElement("array")
        elif isinstance(o, six.string_types):
            self.text(o)
        elif isinstance(o, six.integer_types):
            self.text("%d" % o)
        elif isinstance(o, float):
            self.text("%s" % o)
        elif isinstance(o, bool):
            if o:
                self.text("True")
            else:
                self.text("False")
        elif o is None:
            pass
        else:
            result = self.encoder.encode(o, *args, **kwargs)

            if result is None:
                raise TypeError("%r is not XML serializable" % (o,))

            return self.__encode(result, *args, **kwargs)

    def startElement(self, name, attrs={}):
        self.addIndent()
        self.xml.startElement(name, attrs)
        self.level += 1
        self.doIndent = True

    def endElement(self, name):
        self.level -= 1
        self.addIndent()
        self.xml.endElement(name)
        self.doIndent = True

    def text(self, value):
        self.xml.characters(value)
        self.doIndent = False

    def addIndent(self):
        if self.doIndent:
            self.xml.ignorableWhitespace('\n' + ' ' * self.level)


_registered_encoders = None


def get_registered_encoders():
    """
    Returns a list of registered Web API encoders.
    """
    global _registered_encoders

    if _registered_encoders is None:
        _registered_encoders = []

        encoders = getattr(settings, 'WEB_API_ENCODERS',
                           ['djblets.webapi.encoders.BasicAPIEncoder'])

        for encoder in encoders:
            encoder_path = encoder.split('.')
            if len(encoder_path) > 1:
                encoder_module_name = '.'.join(encoder_path[:-1])
            else:
                encoder_module_name = '.'

            encoder_module = __import__(encoder_module_name, {}, {},
                                        encoder_path[-1])
            encoder_class = getattr(encoder_module, encoder_path[-1])
            _registered_encoders.append(encoder_class())

    return _registered_encoders



from __future__ import unicode_literals

from django.db import models
from django.db.models import F

from djblets.db.fields import CounterField
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin


class CounterFieldMixin(models.Model):
    mixin_counter = CounterField(initializer=lambda o: 0)

    class Meta:
        abstract = True


class CounterFieldModelWithMixin(CounterFieldMixin, models.Model):
    counter = CounterField(initializer=lambda o: 1)


class CounterFieldTestModel(models.Model):
    counter = CounterField(initializer=lambda o: 5)


class CounterFieldInitializerModel(models.Model):
    counter = CounterField(initializer=lambda o: 42)


class CounterFieldInitializerFModel(models.Model):
    INIT_EXPR = F('my_int') + 1

    my_int = models.IntegerField(default=42)

    counter = CounterField(
        initializer=lambda o: CounterFieldInitializerFModel.INIT_EXPR)


class CounterFieldNoInitializerModel(models.Model):
    counter = CounterField()


class CounterFieldTests(TestModelsLoaderMixin, TestCase):
    """Tests for djblets.db.fields.CounterField."""
    tests_app = 'djblets.db.tests'

    def test_no_initializer(self):
        """Testing CounterField without an initializer"""
        model = CounterFieldNoInitializerModel.objects.create()
        self.assertIsNone(model.counter)

    def test_initializer_function(self):
        """Testing CounterField with an initializer function"""
        model = CounterFieldInitializerModel.objects.create()
        self.assertEqual(model.counter, 42)

    def test_initializer_f_existing_instance(self):
        """Testing CounterField with an initializer F() expression
        on existing model instance
        """
        model = CounterFieldInitializerFModel.objects.create()
        model.counter = None
        model.save()

        model = CounterFieldInitializerFModel.objects.get(pk=model.pk)
        self.assertEqual(model.counter, 43)

    def test_initializer_f_new_instance(self):
        """Testing CounterField with an initializer F() expression
        on new model instance
        """
        model = CounterFieldInitializerFModel.objects.create()
        self.assertEqual(model.counter, 0)

    def test_increment(self):
        """Testing CounterField.increment"""
        self._test_increment(expected_value=6, expected_expr=F('counter') + 1)

    def test_increment_by(self):
        """Testing CounterField.increment with increment_by"""
        self._test_increment(expected_value=8, expected_expr=F('counter') + 3,
                             increment_by=3)

    def test_increment_no_reload(self):
        """Testing CounterField.increment with reload_object=False"""
        self._test_increment(expected_value=5, expected_expr=F('counter') + 1,
                             reload_object=False)

    def test_decrement(self):
        """Testing CounterField.decrement"""
        self._test_decrement(expected_value=4, expected_expr=F('counter') - 1)

    def test_decrement_by(self):
        """Testing CounterField.decrement with decrement_by"""
        self._test_decrement(expected_value=2, expected_expr=F('counter') - 3,
                             decrement_by=3)

    def test_decrement_no_reload(self):
        """Testing CounterField.decrement with reload_object=False"""
        self._test_decrement(expected_value=5, expected_expr=F('counter') - 1,
                             reload_object=False)

    def test_reload(self):
        """Testing CounterField.reload"""
        model = CounterFieldTestModel.objects.create()

        model.counter = None
        model.reload_counter()
        self.assertEqual(model.counter, 5)

    def test_save_skips_counters_with_value(self):
        """Testing CounterField with value is skipped during model save"""
        model = CounterFieldTestModel.objects.create()
        model.increment_counter()
        model.counter = 123
        model.save()

        model = CounterFieldTestModel.objects.get(pk=model.pk)
        self.assertEqual(model.counter, 6)

    def test_save_allows_counters_with_none(self):
        """Testing CounterField with None value is allowed during model save"""
        model = CounterFieldTestModel.objects.create()

        # Force bad data in the database.
        CounterFieldTestModel.objects.update(counter=123)
        model.reload_counter()

        self.assertEqual(model.counter, 123)
        model.counter = None
        model.save()

        model = CounterFieldTestModel.objects.get(pk=model.pk)
        self.assertEqual(model.counter, 5)

    def test_save_allows_counters_with_update_fields(self):
        """Testing CounterField with value and update_fields is allowed during
        model save
        """
        model = CounterFieldTestModel.objects.create()
        model.counter = 123
        model.save(update_fields=['counter'])

        model = CounterFieldTestModel.objects.get(pk=model.pk)
        self.assertEqual(model.counter, 123)

    def test_save_with_mixin_counter_field(self):
        """Testing Model.save when the model has a CounterField and inherits
        from an abstract model that has a CounterField
        """
        # This test fails by blowing the stack during ``model.save()``.
        model = CounterFieldModelWithMixin.objects.create()
        model.save()

    def _test_increment(self, expected_value, expected_expr, **kwargs):
        self._test_update_value(expected_value, expected_expr,
                                'increment_counter', **kwargs)

    def _test_decrement(self, expected_value, expected_expr, **kwargs):
        self._test_update_value(expected_value, expected_expr,
                                'decrement_counter', **kwargs)

    def _test_update_value(self, expected_value, expected_expr, func_name,
                           **kwargs):
        model = CounterFieldTestModel.objects.create()

        getattr(model, func_name)(**kwargs)

        self.assertEqual(model.counter, expected_value)



# TODO: Remove this file once we no longer support a version of Django
#       prior to 1.7.
#
# This is Django 1.6.11's django/core/files/locks.py.
#
# Copyright (c) Django Software Foundation and individual contributors.  All
# rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be
#     used to endorse or promote products derived from this software without
#     specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""
Portable file locking utilities.

Based partially on an example by Jonathan Feignberg in the Python
Cookbook [1] (licensed under the Python Software License) and a ctypes port by
Anatoly Techtonik for Roundup [2] (license [3]).

[1] http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/65203
[2] http://sourceforge.net/p/roundup/code/ci/default/tree/roundup/backends/portalocker.py
[3] http://sourceforge.net/p/roundup/code/ci/default/tree/COPYING.txt
"""
import os

__all__ = ('LOCK_EX', 'LOCK_SH', 'LOCK_NB', 'lock', 'unlock')


def _fd(f):
    """Get a filedescriptor from something which could be a file or an fd."""
    return f.fileno() if hasattr(f, 'fileno') else f


if os.name == 'nt':
    import msvcrt
    from ctypes import (sizeof, c_ulong, c_void_p, c_int64,
                        Structure, Union, POINTER, windll, byref)
    from ctypes.wintypes import BOOL, DWORD, HANDLE

    LOCK_SH = 0  # the default
    LOCK_NB = 0x1  # LOCKFILE_FAIL_IMMEDIATELY
    LOCK_EX = 0x2  # LOCKFILE_EXCLUSIVE_LOCK

    # --- Adapted from the pyserial project ---
    # detect size of ULONG_PTR
    if sizeof(c_ulong) != sizeof(c_void_p):
        ULONG_PTR = c_int64
    else:
        ULONG_PTR = c_ulong
    PVOID = c_void_p

    # --- Union inside Structure by stackoverflow:3480240 ---
    class _OFFSET(Structure):
        _fields_ = [
            ('Offset', DWORD),
            ('OffsetHigh', DWORD)]

    class _OFFSET_UNION(Union):
        _anonymous_ = ['_offset']
        _fields_ = [
            ('_offset', _OFFSET),
            ('Pointer', PVOID)]

    class OVERLAPPED(Structure):
        _anonymous_ = ['_offset_union']
        _fields_ = [
            ('Internal', ULONG_PTR),
            ('InternalHigh', ULONG_PTR),
            ('_offset_union', _OFFSET_UNION),
            ('hEvent', HANDLE)]

    LPOVERLAPPED = POINTER(OVERLAPPED)

    # --- Define function prototypes for extra safety ---
    LockFileEx = windll.kernel32.LockFileEx
    LockFileEx.restype = BOOL
    LockFileEx.argtypes = [HANDLE, DWORD, DWORD, DWORD, DWORD, LPOVERLAPPED]
    UnlockFileEx = windll.kernel32.UnlockFileEx
    UnlockFileEx.restype = BOOL
    UnlockFileEx.argtypes = [HANDLE, DWORD, DWORD, DWORD, LPOVERLAPPED]

    def lock(f, flags):
        hfile = msvcrt.get_osfhandle(_fd(f))
        overlapped = OVERLAPPED()
        ret = LockFileEx(hfile, flags, 0, 0, 0xFFFF0000, byref(overlapped))
        return bool(ret)

    def unlock(f):
        hfile = msvcrt.get_osfhandle(_fd(f))
        overlapped = OVERLAPPED()
        ret = UnlockFileEx(hfile, 0, 0, 0xFFFF0000, byref(overlapped))
        return bool(ret)
else:
    try:
        import fcntl
        LOCK_SH = fcntl.LOCK_SH  # shared lock
        LOCK_NB = fcntl.LOCK_NB  # non-blocking
        LOCK_EX = fcntl.LOCK_EX
    except (ImportError, AttributeError):
        # File locking is not supported.
        LOCK_EX = LOCK_SH = LOCK_NB = 0

        # Dummy functions that don't do anything.
        def lock(f, flags):
            # File is not locked
            return False

        def unlock(f):
            # File is unlocked
            return True
    else:
        def lock(f, flags):
            ret = fcntl.flock(_fd(f), flags)
            return (ret == 0)

        def unlock(f):
            ret = fcntl.flock(_fd(f), fcntl.LOCK_UN)
            return (ret == 0)



"""Unit tests for the Root Resource."""

from djblets.extensions.manager import ExtensionManager
from djblets.extensions.resources import ExtensionResource
from djblets.testing.testcases import TestCase
from djblets.webapi.resources import RootResource


class RootResourceTemplateRegistrationTests(TestCase):
    """Unit tests for the (un)registration of templates in RootResource."""

    def setUp(self):
        super(RootResourceTemplateRegistrationTests, self).setUp()

        self.ext_mgr = ExtensionManager('')
        self.ext_res = ExtensionResource(self.ext_mgr)
        self.root_res = RootResource([self.ext_res])
        self.root_res._registered_uri_templates = {
            self.ext_res: {
                'extensions': 'http://localhost:8080/api/extensions/'
            },
            None: {
                'extensions': 'http://localhost:8080/api/extensions/none/'
            },
        }

    def test_register_uri_template_without_relative_resource(self):
        """Testing register_uri_templates without a relative resource"""
        self.root_res.register_uri_template(name='key', relative_path='value')
        actual_result = self.root_res._registered_uri_templates[None]
        self.assertEqual(actual_result, {
            'extensions': 'http://localhost:8080/api/extensions/none/',
            'key': 'value',
        })

    def test_register_uri_template_with_relative_resource(self):
        """Testing register_uri_templates with a relative resource"""
        mock_extension_resource = ExtensionResource(self.ext_mgr)
        self.root_res.register_uri_template(
            name='key',
            relative_path='value',
            relative_resource=mock_extension_resource)
        actual_result = self.root_res._registered_uri_templates[
            mock_extension_resource]
        self.assertEqual(actual_result, {'key': 'value'})

    def test_unregister_uri_template_without_relative_resource(self):
        """Testing unregister_uri_template without a relative resource"""
        self.root_res.unregister_uri_template('extensions')
        self.assertFalse(self.root_res._registered_uri_templates[None])

    def test_unregister_uri_template_with_relative_resource(self):
        """Testing unregister_uri_template with a relative resource"""
        self.root_res.unregister_uri_template('extensions', self.ext_res)
        self.assertFalse(self.root_res._registered_uri_templates[
            self.ext_res])

    def test_register_uri_template_clears_uri_template_cache(self):
        """Testing register_uri_templates clears the URI template cache"""
        self.root_res._uri_templates = {
            'key1': 'value1',
            'key2': 'value2',
        }
        self.root_res.register_uri_template(
            'extension_name', 'some/relative/path/')
        self.assertEqual(self.root_res._uri_templates, {})

    def test_unregister_uri_template_clears_uri_template_cache(self):
        """Testing unregister_uri_templates clears the URI template cache"""
        self.root_res._uri_templates = {
            'key1': 'value1',
            'key2': 'value2',
        }
        self.root_res.unregister_uri_template('extension-1')
        self.assertEqual(self.root_res._uri_templates, {})



from __future__ import unicode_literals

from django.conf.urls import include, url


def dummy_view(request):
    pass


urlpatterns = [
    url(r'^$', dummy_view, name='test-url-name'),
    url(r'^admin/extensions/', include('djblets.extensions.urls')),
    url(r'^auth/', include('djblets.auth.urls')),
]



#
# dates.py -- Date-related utilities.
#
# Copyright (c) 2008-2009  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import unicode_literals

import calendar
from datetime import datetime

from django.db.models import DateField
from django.utils import six
from django.utils.timezone import utc


def http_date(timestamp):
    """
    A wrapper around Django's http_date that accepts DateFields and
    datetime objects directly.
    """
    from django.utils.http import http_date

    if isinstance(timestamp, (DateField, datetime)):
        return http_date(calendar.timegm(timestamp.timetuple()))
    elif isinstance(timestamp, six.string_types):
        return timestamp
    else:
        return http_date(timestamp)


def get_latest_timestamp(timestamps):
    """
    Returns the latest timestamp in a list of timestamps.
    """
    latest = None

    for timestamp in timestamps:
        if latest is None or timestamp > latest:
            latest = timestamp

    return latest


def get_tz_aware_utcnow():
    """Returns a UTC aware datetime object"""
    return datetime.utcnow().replace(tzinfo=utc)



"""Database models for storing site configuration."""

from __future__ import unicode_literals

from django.contrib.sites.models import Site
from django.db import models
from django.utils import six
from django.utils.encoding import python_2_unicode_compatible

from djblets.cache.synchronizer import GenerationSynchronizer
from djblets.db.fields import JSONField
from djblets.siteconfig.managers import SiteConfigurationManager


_GLOBAL_DEFAULTS = {}
_DEFAULTS = {}


class SiteConfigSettingsWrapper(object):
    """Wraps the settings for a SiteConfiguration.

    This is used by the context processor for templates to wrap accessing
    settings data, properly returning defaults.
    """

    def __init__(self, siteconfig):
        """Initialize the wrapper.

        Args:
            siteconfig (SiteConfiguration):
                The site configuration to wrap.
        """
        self.siteconfig = siteconfig

    def __getattr__(self, name):
        """Return an attribute from the site configuration.

        If the attribute is not present in the site configuration's settings,
        the registered default will be returned.

        Args:
            name (unicode):
                The name of the attribute.

        Returns:
            unicode:
            The resulting value from the site configuration or the default.
        """
        return self.siteconfig.get(name)


@python_2_unicode_compatible
class SiteConfiguration(models.Model):
    """Stored version and settings data for a Django site.

    This stores dynamic settings for a site, along with version information,
    allowing the application to alter and apply/synchronize settings across
    threads, processes, and servers without restarting the server.

    Consumers should not create or fetch their own instance of this class
    through standard Django query functions. Instead, they should use
    :py:meth:`SiteConfiguration.objects.get_current()
    <djblets.siteconfig.managers.SiteConfigurationManager.get_current>`
    instead. See the documentation for that method for details on how to safely
    look up and use site configuration.
    """

    site = models.ForeignKey(Site,
                             related_name='config',
                             on_delete=models.CASCADE)
    version = models.CharField(max_length=20)

    #: A JSON dictionary field of settings stored for a site.
    settings = JSONField()

    objects = SiteConfigurationManager()

    @classmethod
    def add_global_defaults(cls, defaults_dict):
        """Add a dictionary of global defaults for settings.

        These defaults will be used when calling :py:meth:`get` for any setting
        not stored. Defaults registered for a specific site configuration take
        precedent over global defaults.

        Args:
            default_dict (dict):
                A dictionary of defaults, mapping siteconfig settings keys to
                JSON-serializable values.
        """
        _GLOBAL_DEFAULTS.update(defaults_dict)

    @classmethod
    def add_global_default(cls, key, default_value):
        """Add a global default value for a settings key.

        The default will be used when calling :py:meth:`get` for this key,
        if a value is not stored. Defaults registered for a specific site
        configuration take precedent over global defaults.

        Args:
            key (unicode):
                The settings key to set the default for.

            default_value (object):
                The value to set as the default.
        """
        cls.add_global_defaults({key: default_value})

    @classmethod
    def remove_global_default(self, key):
        """Remove a global default value for a settings key.

        Args:
            key (unicode):
                The settings key to remove the default for.
        """
        _GLOBAL_DEFAULTS.pop(key)

    @classmethod
    def clear_global_defaults(self):
        """Clear all default values for this site configuration.

        This will clear only global defaults. This will not affect defaults
        registered on specific site configurations.
        """
        _GLOBAL_DEFAULTS.clear()

    @classmethod
    def get_global_defaults(cls):
        """Return all global defaults for settings.

        Returns:
            dict:
            A dictionary of all registered global defaults for settings.
        """
        return _GLOBAL_DEFAULTS

    def __init__(self, *args, **kwargs):
        """Initialize the site configuration.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent constructor.

            **kwargs (dict):
                Keyword arguments to pass to the parent constructor.
        """
        super(SiteConfiguration, self).__init__(*args, **kwargs)

        # Optimistically try to set the Site to the current site instance,
        # which either is cached now or soon will be. That way, we avoid
        # a lookup on the relation later.
        cur_site = Site.objects.get_current()

        if cur_site.pk == self.site_id:
            self.site = cur_site

        # Begin managing the synchronization of settings between all
        # SiteConfigurations.
        self._gen_sync = GenerationSynchronizer(
            '%s:siteconfig:%s:generation' % (self.site.domain, self.pk))

        self.settings_wrapper = SiteConfigSettingsWrapper(self)

    def get(self, key, default=None):
        """Return the value for a setting.

        If the setting is not found, the default value will be returned. This
        is represented by the default parameter, if passed in, or a global
        default (from :py:meth:`add_default`) if set.

        If no default is available, ``None`` will be returned.

        Args:
            key (unicode):
                The site configuration settings key.

            default (object, optional):
                The default value to return. If not provided, the registered
                default will be returned.

        Returns:
            object:
            The resulting value.
        """
        if default is None:
            try:
                default = _DEFAULTS[self.pk][key]
            except KeyError:
                default = _GLOBAL_DEFAULTS.get(key)

        return self.settings.get(key, default)

    def set(self, key, value):
        """Set a value for a setting.

        The setting will be stored locally until the model is saved, at which
        point it will be synchronized with other processes/servers.

        Args:
            key (unicode):
                The key for the setting.

            value (object):
                The JSON-serializable object to store.
        """
        self.settings[key] = value

    def add_defaults(self, defaults_dict):
        """Add a dictionary of defaults for settings.

        These defaults will be used when calling :py:meth:`get` for any setting
        not stored.  These will only be registered for this site configuration,
        and will not be registered for global defaults.

        Args:
            default_dict (dict):
                A dictionary of defaults, mapping siteconfig settings keys to
                JSON-serializable values.
        """
        _DEFAULTS.setdefault(self.pk, {}).update(defaults_dict)

    def add_default(self, key, default_value):
        """Add a default value for a settings key.

        The default will be used when calling :py:meth:`get` for this key,
        if a value is not stored. This will only be registered for this site
        configuration, and will not be registered for global defaults.

        Args:
            key (unicode):
                The settings key to set the default for.

            default_value (object):
                The value to set as the default.
        """
        self.add_defaults({key: default_value})

    def remove_default(self, key):
        """Remove a default value on this site configuration.

        This will remove only defaults registered on this site configuration.
        This does not affect global defaults.

        Args:
            key (unicode):
                The settings key to remove the default for.
        """
        try:
            del _DEFAULTS[self.pk][key]
        except KeyError:
            pass

    def clear_defaults(self):
        """Clear all default values for this site configuration.

        This will clear only defaults registered on this site configuration.
        This does not affect global defaults.
        """
        _DEFAULTS[self.pk] = {}

    def get_defaults(self):
        """Return all defaults for this site configuration.

        This will return only defaults registered on this site configuration.
        The result does not include global defaults.

        Returns:
            dict:
            A dictionary of all registered defaults for settings.
        """
        return _DEFAULTS.get(self.pk, {})

    def is_expired(self):
        """Return whether or not this SiteConfiguration is expired.

        If the configuration is expired, it will need to be reloaded before
        accessing any settings.

        Returns:
            bool:
            Whether or not the current state is expired.
        """
        return self._gen_sync.is_expired()

    def save(self, clear_caches=True, **kwargs):
        """Save the site configuration to the database.

        By default, saving will clear the caches across all processes/servers
        using this site configuration, causing them to be re-fetched on the
        next request.

        Args:
            clear_caches (bool, optional):
                Whether to clear the caches. This is ``True`` by default.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent method.
        """
        self._gen_sync.mark_updated()

        if clear_caches:
            # The cached siteconfig might be stale now. We'll want a refresh.
            # Also refresh the Site cache, since callers may get this from
            # Site.config.
            SiteConfiguration.objects.clear_cache()
            Site.objects.clear_cache()

        super(SiteConfiguration, self).save(**kwargs)

    def __str__(self):
        """Return a string version of the site configuration.

        The returned string will list the associated site's domain and the
        stored application version.

        Returns:
            unicode:
            The string representation of the site configuration.
        """
        return "%s (version %s)" % (six.text_type(self.site), self.version)

    class Meta:
        # Djblets 0.9+ sets an app label of "djblets_siteconfig" on
        # Django 1.7+, which would affect the table name. We need to retain
        # the old name for backwards-compatibility.
        db_table = 'siteconfig_siteconfiguration'



"""Common forms for authentication tasks.

These forms can be used either directly or as a base for more specific forms.
They're designed to take some of the complexity out of creating
authentication-related forms not otherwise provided by Django.
"""

from __future__ import unicode_literals

from django import forms
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db import transaction
from django.utils.translation import ugettext_lazy as _


class RegistrationForm(forms.Form):
    """A standard registration form collecting basic account information.

    This form prompts the user for a username, a password (and a confirmation
    on that password), e-mail address, first name, and last name. It then
    validates these, attempting to create a :py:class:`User`.

    This class can be extended by subclasses to provide additional fields.
    """

    username = forms.RegexField(
        r"^[a-zA-Z0-9_\-\.]*$",
        max_length=30,
        error_messages={
            'invalid': _('Only A-Z, 0-9, "_", "-", and "." are allowed.'),
        })
    password1 = forms.CharField(label=_('Password'),
                                min_length=5,
                                widget=forms.PasswordInput)
    password2 = forms.CharField(label=_('Password (confirm)'),
                                widget=forms.PasswordInput)
    email = forms.EmailField(label=_('E-mail address'))
    first_name = forms.CharField(max_length=30, required=False)
    last_name = forms.CharField(max_length=30, required=False)

    def __init__(self, request=None, *args, **kwargs):
        """Initialize the form.

        Args:
            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments to pass to the parent class.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        super(RegistrationForm, self).__init__(*args, **kwargs)
        self.request = request

    def clean_password2(self):
        """Validate that the two supplied passwords match.

        If they do not match, validation will fail, and an error will be
        supplied to the user.

        Returns:
            unicode:
            The password supplied on the form.

        Raises:
            django.core.exceptions.ValidationError:
                If the passwords do not match.
        """
        formdata = self.cleaned_data

        if ('password1' in formdata and
            formdata['password1'] != formdata['password2']):
            raise ValidationError(_('Passwords must match'))

        return formdata['password2']

    def save(self):
        """Save the form, creating a user if validation passes.

        The user will be created with the provided username, e-mail address,
        password, and full name. If there are failures in creating this user,
        or there's an existing user with the given name, an error will be
        raised. Creation always occurs in its own transaction, ensuring it's
        immediately saved and preventing potential lookup problems when there's
        a conflict.

        Subclasses that want to override this can call the parent's
        py:meth:`save` and modify the resulting user, if ``None`` is not
        returned.

        Returns:
            django.contrib.auth.models.User:
            The newly-created user, or ``None`` if there was a conflict when
            creating it.

        Raises:
            Exception:
                An unexpected exception occurred.
        """
        if not self.errors:
            username = self.cleaned_data['username']

            try:
                with transaction.atomic():
                    return User.objects.create_user(
                        username=username,
                        password=self.cleaned_data['password1'],
                        email=self.cleaned_data['email'],
                        first_name=self.cleaned_data['first_name'],
                        last_name=self.cleaned_data['last_name'])
            except:
                # We check for duplicate users here instead of clean, since
                # it's possible that two users could race for a name.
                if User.objects.filter(username=username).exists():
                    self.errors['username'] = self.error_class(
                        [_('Sorry, this username is taken.')])
                else:
                    raise

        return None



#
# tests.py -- Unit tests for classes in djblets.datagrid
#
# Copyright (c) 2007-2008  Christian Hammond
# Copyright (c) 2007-2008  David Trowbridge
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

from datetime import datetime, timedelta

from django.conf import settings
from django.contrib.auth.models import Group, User
from django.http import HttpRequest
from django.test.client import RequestFactory
from django.utils.encoding import force_text
from django.utils.safestring import SafeText
from django.utils.six.moves import range
from kgb import SpyAgency

from djblets.datagrid.grids import (CheckboxColumn, Column, DataGrid,
                                    DateTimeSinceColumn, StatefulColumn,
                                    logger)
from djblets.testing.testcases import TestCase
from djblets.util.dates import get_tz_aware_utcnow


class GroupDataGrid(DataGrid):
    objid = Column('ID', link=True, sortable=True, field_name='id')
    name = Column('Group Name', link=True, sortable=True, expand=True)

    def __init__(self, request):
        super(GroupDataGrid, self).__init__(request=request,
                                            queryset=Group.objects.all(),
                                            title='All Groups')
        self.default_sort = []
        self.default_columns = ['objid', 'name']


class CheckboxColumnTests(TestCase):
    """Unit tests for djblets.datagrid.grids.CheckboxColumn."""

    def test_initial_state(self):
        """Testing CheckboxColumn initial state"""
        column = CheckboxColumn(checkbox_name='my_checkbox&',
                                detailed_label='<Select Rows>')

        self.assertHTMLEqual(
            column.label,
            '<input class="datagrid-header-checkbox" '
            ' type="checkbox" data-checkbox-name="my_checkbox&amp;">')
        self.assertHTMLEqual(
            column.detailed_label_html,
            '<input type="checkbox"> &lt;Select Rows&gt;')
        self.assertEqual(column.detailed_label, '<Select Rows>')
        self.assertEqual(column.checkbox_name, 'my_checkbox&')

    def test_render_data_with_selected(self):
        """Testing CheckboxColumn.render_data with selected object"""
        group = Group.objects.create(name='test')

        column = CheckboxColumn(checkbox_name='my_checkbox&',
                                detailed_label='<Select Rows>')
        column.is_selected = lambda *args, **kwargs: True

        self.assertHTMLEqual(
            column.render_data(state=None, obj=group),
            '<input type="checkbox" data-object-id="1"'
            ' data-checkbox-name="my_checkbox&amp;" checked="true">')

    def test_render_data_with_unselected(self):
        """Testing CheckboxColumn.render_data with unselected object"""
        group = Group.objects.create(name='test')

        column = CheckboxColumn(checkbox_name='my_checkbox&',
                                detailed_label='<Select Rows>')

        self.assertHTMLEqual(
            column.render_data(state=None, obj=group),
            '<input type="checkbox" data-object-id="1"'
            ' data-checkbox-name="my_checkbox&amp;">')


class DateTimeSinceColumnTests(TestCase):
    """Unit tests for djblets.datagrid.grids.DateTimeSinceColumn."""

    def test_render_data(self):
        """Testing DateTimeSinceColumn.render_data"""
        class DummyObj:
            time = None

        column = DateTimeSinceColumn('Test', field_name='time')
        state = StatefulColumn(None, column)

        if settings.USE_TZ:
            now = get_tz_aware_utcnow()
        else:
            now = datetime.now()

        obj = DummyObj()
        obj.time = now
        self.assertEqual(column.render_data(state, obj), '0\xa0minutes ago')

        obj.time = now - timedelta(days=5)
        self.assertEqual(column.render_data(state, obj), '5\xa0days ago')

        obj.time = now - timedelta(days=7)
        self.assertEqual(column.render_data(state, obj), '1\xa0week ago')


class DataGridTests(SpyAgency, TestCase):
    """Unit tests for djblets.datagrid.grids.DataGrid."""

    def setUp(self):
        super(DataGridTests, self).setUp()

        Group.objects.bulk_create(
            Group(name='Group %02d' % i)
            for i in range(1, 100)
        )

        self.user = User(username='testuser')
        self.request = HttpRequest()
        self.request.user = self.user
        self.datagrid = GroupDataGrid(self.request)

    def test_render_listview(self):
        """Testing DataGrid.render_listview"""
        result = self.datagrid.render_listview()

        self.assertIsInstance(result, SafeText)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)
        self.assertIn('<div class="datagrid-main">', result)
        self.assertInHTML('<td colspan="2">Group 01</td>', result)
        self.assertInHTML('<td colspan="2">Group 50</td>', result)
        self.assertIn('<div class="paginator">', result)
        self.assertIn('<div class="datagrid-menu', result)

    def test_render_listview_to_response(self):
        """Testing DataGrid.render_listview_to_response"""
        response = self.datagrid.render_listview_to_response()
        self.assertEqual(response.status_code, 200)

        content = force_text(response.content)

        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">',
                      content)
        self.assertIn('<div class="datagrid-main">', content)
        self.assertInHTML('<td colspan="2">Group 01</td>', content)
        self.assertInHTML('<td colspan="2">Group 50</td>', content)
        self.assertIn('<div class="paginator">', content)
        self.assertIn('<div class="datagrid-menu', content)

    def test_render_paginator(self):
        """Testing DataGrid.render_paginator"""
        self.datagrid.load_state()
        content = self.datagrid.render_paginator()

        self.assertHTMLEqual(
            content,
            '<div class="paginator">'
            ' <span class="current-page">1</span>'
            ' <a href="?page=2" title="Page 2">2</a>'
            ' <a href="?page=2" title="Next Page">&gt;</a>'
            ' <span class="page-count">2 pages&nbsp;</span>'
            '</div>')

    def test_load_state_with_sort_ascending(self):
        """Testing DataGrid.load_state with ascending sort"""
        self.request.GET['sort'] = 'name,objid'
        self.datagrid.load_state()

        self.assertEqual(self.datagrid.sort_list, ['name', 'objid'])
        self.assertEqual(len(self.datagrid.rows), self.datagrid.paginate_by)
        self.assertEqual(self.datagrid.rows[0]['object'].name, 'Group 01')
        self.assertEqual(self.datagrid.rows[1]['object'].name, 'Group 02')
        self.assertEqual(self.datagrid.rows[2]['object'].name, 'Group 03')

        # Exercise the code paths when rendering.
        result = self.datagrid.render_listview()
        self.assertIsInstance(result, SafeText)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)

    def test_load_state_with_sort_descending(self):
        """Testing DataGrid.load_state with descending sort"""
        self.request.GET['sort'] = '-name'
        self.datagrid.load_state()

        self.assertEqual(self.datagrid.sort_list, ['-name'])
        self.assertEqual(len(self.datagrid.rows), self.datagrid.paginate_by)
        self.assertEqual(self.datagrid.rows[0]['object'].name, 'Group 99')
        self.assertEqual(self.datagrid.rows[1]['object'].name, 'Group 98')
        self.assertEqual(self.datagrid.rows[2]['object'].name, 'Group 97')

        # Exercise the code paths when rendering.
        result = self.datagrid.render_listview()
        self.assertIsInstance(result, SafeText)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)

    def test_load_state_with_custom_column_orders(self):
        """Testing DataGrid.load_state with custom column orders"""
        self.request.GET['columns'] = 'objid'
        self.datagrid.load_state()

        self.assertEqual(len(self.datagrid.rows), self.datagrid.paginate_by)
        self.assertEqual(len(self.datagrid.rows[0]['cells']), 1)

        # Exercise the code paths when rendering.
        result = self.datagrid.render_listview()
        self.assertIsInstance(result, SafeText)
        self.assertIn('<div class="datagrid-wrapper" id="datagrid-0">', result)

    def test_precompute_objects_with_unsortable_column_ascending(self):
        """Testing DataGrid.precompute_objects with improper sort key
        in ascending order
        """
        self.request.GET['sort'] = 'garbagetrash'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, [])

    def test_precompute_objects_with_unsortable_column_descending(self):
        """Testing DataGrid.load_state with improper sort key in
        descending order
        """
        self.request.GET['sort'] = '-garbagetrash'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, [])

    def test_precompute_objects_with_mixed_column_ascending(self):
        """Testing DataGrid.load_state with one proper and one improper sort
        key in ascending order
        """
        self.request.GET['sort'] = 'garbagetrash,name'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, ['name'])

    def test_precompute_objects_with_mixed_column_descending(self):
        """Testing DataGrid.load_state with one proper and one improper sort
        key in descending order
        """
        self.request.GET['sort'] = '-garbagetrash,-name'
        self.datagrid.load_state()
        self.datagrid.precompute_objects()

        self.assertEqual(self.datagrid.sort_list, ['-name'])

    def test_post_process_queryset_with_select_related(self):
        """Testing DataGrid.post_process_queryset with chained select_related
        calls
        """
        class Column1(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo')

        class Column2(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo__bar')

        class Column3(Column):
            def augment_queryset(self, state, queryset):
                return queryset.select_related('foo', 'foo__abc')

        class TestDataGrid(DataGrid):
            col1 = Column1()
            col2 = Column2()
            col3 = Column3()

            def __init__(self, *args, **kwargs):
                super(TestDataGrid, self).__init__(*args, **kwargs)

                self.default_columns = ['col1', 'col2', 'col3']
                self.default_sort = []

            def post_process_queryset(self, queryset):
                return super(TestDataGrid, self).post_process_queryset(
                    queryset.select_related('baz'))

        grid = TestDataGrid(self.request, queryset=Group.objects.all())
        grid.columns = [
            grid.get_stateful_column(grid.get_column(name))
            for name in grid.default_columns
        ]

        queryset = grid.post_process_queryset(grid.queryset)

        self.assertEqual(
            queryset.query.select_related,
            {
                'foo': {
                    'bar': {},
                    'abc': {},
                },
                'baz': {},
            })

    def test_post_process_queryset_sandboxes_errors(self):
        """Testing DataGrid.post_process_queryset with column.augment_queryset
        raises exception
        """
        class BadColumn(Column):
            def augment_queryset(self, state, queryset):
                raise Exception

        column = BadColumn(id='test')
        self.datagrid.add_column(column)

        self.spy_on(column.augment_queryset)
        self.spy_on(logger.exception)

        try:
            self.datagrid.columns = [
                self.datagrid.get_stateful_column(column),
            ]
            self.datagrid.post_process_queryset(queryset=Group.objects.all())
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.augment_queryset.called)
        self.assertIn('Error when calling augment_queryset for DataGrid '
                      'Column',
                      logger.exception.last_call.args[0])

    def test_get_stateful_column_sandboxes_errors(self):
        """Testing DataGrid.get_stateful_column when column.setup_state
        raises exception
        """
        class BadColumn(Column):
            def setup_state(self, state):
                raise Exception

        column = BadColumn(id='test')
        self.datagrid.add_column(column)

        self.spy_on(column.setup_state)
        self.spy_on(logger.exception)

        try:
            self.datagrid.get_stateful_column(column=column)
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.setup_state.called)
        self.assertIn('Error when calling setup_state for DataGrid Column',
                      logger.exception.last_call.args[0])

    def test_get_sort_field_columns_sandboxes_errors(self):
        """Testing DataGrid.get_sort_field_columns when column.get_sort_field
        raises exception
        """
        class BadColumn(Column):
            def get_sort_field(self, state):
                raise Exception

        column = BadColumn(id='test', sortable=True)
        self.datagrid.add_column(column)

        self.datagrid.sort_list = ['test']
        self.datagrid.default_columns = ['objid', 'test']

        self.spy_on(column.get_sort_field)
        self.spy_on(logger.exception)

        try:
            self.datagrid.precompute_objects()
        finally:
            self.datagrid.remove_column(column)

        self.assertTrue(column.get_sort_field.called)
        self.assertIn('Error when calling get_sort_field for DataGrid Column',
                      logger.exception.last_call.args[0])


class ColumnTests(SpyAgency, TestCase):
    """Unit tests for djblets.datagrid.grids.Column."""

    def test_render_cell_sandboxes_errors(self):
        """Testing Column.render_cell when column.render_data raises exception
        """
        class BadColumn(Column):
            def render_data(self, state, obj):
                raise Exception

        request = RequestFactory().request()
        datagrid = DataGrid(request=request,
                            queryset=Group.objects.all())

        column = BadColumn(id='test')
        self.spy_on(column.render_data)
        self.spy_on(logger.exception)

        rendered = column.render_cell(state=StatefulColumn(datagrid=datagrid,
                                                           column=column),
                                      obj=None,
                                      render_context=None)
        self.assertHTMLEqual(rendered, '<td></td>')

        self.assertTrue(column.render_data.called)
        self.assertIn('Error when calling render_data for DataGrid Column',
                      logger.exception.last_call.args[0])



"""Unit tests for djblets.privacy.consent.forms.MultiConsentRequirementsField.
"""

from __future__ import unicode_literals

from django.contrib.auth.models import User
from django.core.exceptions import ValidationError

from djblets.privacy.consent import (BaseConsentRequirement, Consent,
                                     get_consent_tracker)
from djblets.privacy.consent.forms import MultiConsentRequirementsField
from djblets.privacy.tests.testcases import ConsentTestCase


class MyConsentRequirement1(BaseConsentRequirement):
    requirement_id = 'my-requirement-1'
    name = 'My Requirement 1'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'


class MyConsentRequirement2(BaseConsentRequirement):
    requirement_id = 'my-requirement-2'
    name = 'My Requirement 2'
    summary = 'We would also like this'
    intent_description = 'We need this for dancing.'
    data_use_description = 'Dancing all the things.'


class MultiConsentRequirementsFieldTests(ConsentTestCase):
    """Unit tests for MultiConsentRequirementsField."""

    def setUp(self):
        super(MultiConsentRequirementsFieldTests, self).setUp()

        self.consent_requirement_1 = MyConsentRequirement1()
        self.consent_requirement_2 = MyConsentRequirement2()

        self.field = MultiConsentRequirementsField(
            consent_requirements=[
                self.consent_requirement_1,
                self.consent_requirement_2,
            ],
            consent_source='https://example.com/consent/',
            extra_consent_data={
                'test': True,
            })

        self.user = User.objects.create(username='test-user')

    def test_init_with_user_and_no_existing_consent(self):
        """Testing MultiConsentRequirementsField.__init__ with user and no
        existing consent data
        """
        self.field.set_initial_from_user(self.user)
        self.assertEqual(self.field.initial, [Consent.UNSET, Consent.UNSET])

    def test_init_with_user_and_existing_consent(self):
        """Testing MultiConsentRequirementsField.__init__ with user and
        existing consent data
        """
        tracker = get_consent_tracker()

        tracker.record_consent_data(
            self.user,
            self.consent_requirement_1.build_consent_data(granted=True))
        tracker.record_consent_data(
            self.user,
            self.consent_requirement_2.build_consent_data(granted=False))

        self.field.set_initial_from_user(self.user)
        self.assertEqual(self.field.initial, [Consent.GRANTED, Consent.DENIED])

    def test_set_initial_from_user(self):
        """Testing MultiConsentRequirementsField.set_initial_from_user"""
        get_consent_tracker().record_consent_data(
            self.user,
            self.consent_requirement_2.build_consent_data(granted=True))

        self.field.set_initial_from_user(self.user)
        self.assertEqual(self.field.initial, [Consent.UNSET, Consent.GRANTED])

    def test_prepare_value(self):
        """Testing MultiConsentRequirementsField.prepare_value"""
        self.assertEqual(
            self.field.prepare_value([Consent.GRANTED, Consent.UNSET]),
            ['allow', None])

    def test_clean_with_valid_values(self):
        """Testing MultiConsentRequirementsField.clean with valid values"""
        values = self.field.clean(['allow', 'block'])

        self.assertIsNotNone(values)
        self.assertEqual(len(values), 2)

        consent_data = values[0]
        self.assertEqual(consent_data.requirement_id, 'my-requirement-1')
        self.assertEqual(consent_data.source, 'https://example.com/consent/')
        self.assertTrue(consent_data.granted)
        self.assertEqual(
            consent_data.extra_data,
            {
                'test': True,
            })

        consent_data = values[1]
        self.assertEqual(consent_data.requirement_id, 'my-requirement-2')
        self.assertEqual(consent_data.source, 'https://example.com/consent/')
        self.assertFalse(consent_data.granted)
        self.assertEqual(
            consent_data.extra_data,
            {
                'test': True,
            })

        # Make sure the timestamps are identical.
        self.assertEqual(values[0].timestamp, values[1].timestamp)

    def test_clean_with_unset(self):
        """Testing MultiConsentRequirementsField.clean with unset"""
        message = 'You must choose Allow or Block for all options to continue.'

        with self.assertRaisesMessage(ValidationError, message):
            self.field.clean(['allow', None])



from __future__ import unicode_literals

from django.conf.urls import url

from djblets.auth.views import register

urlpatterns = [
    url(r'register/', register, kwargs={
        'next_page': 'test',
    }, name='register')
]



#
# views.py -- Views for the log app
#
# Copyright (c) 2009  Christian Hammond
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from __future__ import unicode_literals

import calendar
import codecs
import datetime
import logging
import os
import re
import time

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404
from django.template.context import RequestContext
from django.utils import six
from django.utils.six.moves.urllib.parse import urlencode
from django.utils.translation import ugettext_lazy as _

from djblets.util.compat.django.shortcuts import render


LEVELS = (
    (logging.DEBUG, 'debug', _('Debug')),
    (logging.INFO, 'info', _('Info')),
    (logging.WARNING, 'warning', _('Warning')),
    (logging.ERROR, 'error', _('Error')),
    (logging.CRITICAL, 'critical', _('Critical')),
)


# Matches the default timestamp format in the logging module.
TIMESTAMP_FMT = '%Y-%m-%d %H:%M:%S'

LOG_LINE_RE = re.compile(
    r'(?P<timestamp>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d{3}) - '
    r'(?P<level>DEBUG|INFO|WARNING|ERROR|CRITICAL) - '
    r'(?P<message>.*)')


def parse_timestamp(format, timestamp_str):
    """Utility function to parse a timestamp into a datetime.datetime.

    Python 2.5 and up have datetime.strptime, but Python 2.4 does not,
    so we roll our own as per the documentation.

    If passed a timestamp_str of None, we will return None as a convenience.
    """
    if not timestamp_str:
        return None

    return datetime.datetime(*time.strptime(timestamp_str, format)[0:6])


def build_query_string(request, params):
    """Builds a query string that includes the specified parameters along
    with those that were passed to the page.

    params is a dictionary.
    """
    query_parts = []

    for key, value in six.iteritems(request.GET):
        if key not in params:
            query_parts.append(urlencode({
                key: value
            }))

    for key, value in six.iteritems(params):
        if value is not None:
            query_parts.append(urlencode({
                key: value
            }))

    return '?' + '&'.join(query_parts)


def iter_log_lines(from_timestamp, to_timestamp, requested_levels):
    """Generator that iterates over lines in a log file, yielding the
    yielding information about the lines."""
    log_filename = os.path.join(settings.LOGGING_DIRECTORY,
                                settings.LOGGING_NAME + '.log')

    line_info = None

    try:
        fp = codecs.open(log_filename, encoding='utf-8')
    except IOError:
        # We'd log this, but it'd do very little good in practice.
        # It would only appear on the console when using the development
        # server, but production users would never see anything. So,
        # just return gracefully. We'll show an empty log, which is
        # about accurate.
        return

    for line in fp:
        line = line.rstrip()

        m = LOG_LINE_RE.match(line)

        if m:
            if line_info:
                # We have a fully-formed log line and this new line isn't
                # part of it, so yield it now.
                yield line_info
                line_info = None

            timestamp_str = m.group('timestamp')
            level = m.group('level')
            message = m.group('message')

            if not requested_levels or level.lower() in requested_levels:
                timestamp = parse_timestamp(TIMESTAMP_FMT,
                                            timestamp_str.split(',')[0])

                timestamp_date = timestamp.date()

                if ((from_timestamp and from_timestamp > timestamp_date) or
                    (to_timestamp and to_timestamp < timestamp_date)):
                    continue

                line_info = (timestamp, level, message)
        elif line_info:
            line_info = (line_info[0],
                         line_info[1],
                         line_info[2] + "\n" + line)

    if line_info:
        yield line_info

    fp.close()


def get_log_filtersets(request, requested_levels,
                       from_timestamp, to_timestamp):
    """Returns the filtersets that will be used in the log view."""
    logger = logging.getLogger('')
    level_filters = [
        {
            'name': _('All'),
            'url': build_query_string(request, {'levels': None}),
            'selected': len(requested_levels) == 0,
        }
    ] + [
        {
            'name': label_name,
            'url': build_query_string(request, {'levels': level_name}),
            'selected': level_name in requested_levels,
        }
        for level_id, level_name, label_name in LEVELS
        if logger.isEnabledFor(level_id)
    ]

    from_timestamp_str = request.GET.get('from', None)
    to_timestamp_str = request.GET.get('to', None)
    today = datetime.date.today()
    today_str = today.strftime('%Y-%m-%d')
    one_week_ago = today - datetime.timedelta(days=7)
    one_week_ago_str = one_week_ago.strftime('%Y-%m-%d')
    month_range = calendar.monthrange(today.year, today.month)
    this_month_begin_str = today.strftime('%Y-%m-01')
    this_month_end_str = today.strftime('%Y-%m-') + str(month_range[1])

    date_filters = [
        {
            'name': _('Any date'),
            'url': build_query_string(request, {
                'from': None,
                'to': None,
            }),
            'selected': (from_timestamp_str is None and
                         to_timestamp_str is None),
        },
        {
            'name': _('Today'),
            'url': build_query_string(request, {
                'from': today_str,
                'to': today_str,
            }),
            'selected': (from_timestamp_str == today_str and
                         to_timestamp_str == today_str),
        },
        {
            'name': _('Past 7 days'),
            'url': build_query_string(request, {
                'from': one_week_ago_str,
                'to': today_str,
            }),
            'selected': (from_timestamp_str == one_week_ago_str and
                         to_timestamp_str == today_str),
        },
        {
            'name': _('This month'),
            'url': build_query_string(request, {
                'from': this_month_begin_str,
                'to': this_month_end_str,
            }),
            'selected': (from_timestamp_str == this_month_begin_str and
                         to_timestamp_str == this_month_end_str),
        },
    ]

    return (
        (_("By date"), date_filters),
        (_("By level"), level_filters),
    )


@staff_member_required
def server_log(request, template_name='log/log.html'):
    """Displays the server log."""

    # First check if logging is even configured. If it's not, just return
    # a 404.
    if (not getattr(settings, "LOGGING_ENABLED", False) or
        not getattr(settings, "LOGGING_DIRECTORY", None)):
        raise Http404()

    requested_levels = []

    # Get the list of levels to show.
    if 'levels' in request.GET:
        requested_levels = request.GET.get('levels').split(',')

    # Get the timestamp ranges.
    from_timestamp = parse_timestamp('%Y-%m-%d', request.GET.get('from'))
    to_timestamp = parse_timestamp('%Y-%m-%d', request.GET.get('to'))

    if from_timestamp:
        from_timestamp = from_timestamp.date()

    if to_timestamp:
        to_timestamp = to_timestamp.date()

    # Get the filters to show.
    filtersets = get_log_filtersets(request, requested_levels,
                                    from_timestamp, to_timestamp)

    # Grab the lines from the log file.
    log_lines = iter_log_lines(from_timestamp, to_timestamp, requested_levels)

    # Figure out the sorting
    sort_type = request.GET.get('sort', 'asc')

    if sort_type == 'asc':
        reverse_sort_type = 'desc'
    else:
        reverse_sort_type = 'asc'
        log_lines = reversed(list(log_lines))

    return render(request, template_name, {
        'log_lines': log_lines,
        'filtersets': filtersets,
        'sort_url': build_query_string(request, {'sort': reverse_sort_type}),
        'sort_type': sort_type,
    })



"""Unit tests for djblets.extensions.hooks.ExtensionHook."""

from __future__ import unicode_literals

from django.utils import six

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.testing.testcases import TestCase


@six.add_metaclass(ExtensionHookPoint)
class TestExtensionHook(ExtensionHook):
    pass


class TestExtension(Extension):
    pass


class ExtensionHookTest(ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.ExtensionHook."""

    def setUp(self):
        super(ExtensionHookTest, self).setUp()

        self.extension = self.setup_extension(TestExtension)

    def test_init_hook_states(self):
        """Testing ExtensionHook enabling hook states"""
        @six.add_metaclass(ExtensionHookPoint)
        class InitTestHook(ExtensionHook):
            def initialize(hook):
                self.assertFalse(hook.initialized)
                self.assertEqual(hook.hook_state,
                                 ExtensionHook.HOOK_STATE_ENABLING)

        hook = InitTestHook(self.extension)
        self.assertEqual(hook.hook_state, ExtensionHook.HOOK_STATE_ENABLED)
        self.assertTrue(hook.initialized)

    def test_disable_hook_states(self):
        """Testing ExtensionHook disabling states"""
        @six.add_metaclass(ExtensionHookPoint)
        class InitTestHook(ExtensionHook):
            def shutdown(hook):
                self.assertFalse(hook.initialized)
                self.assertEqual(hook.hook_state,
                                 ExtensionHook.HOOK_STATE_DISABLING)

        hook = InitTestHook(self.extension)
        hook.disable_hook()
        self.assertEqual(hook.hook_state, ExtensionHook.HOOK_STATE_DISABLED)
        self.assertFalse(hook.initialized)

    def test_init_with_start_enabled_false(self):
        """Testing ExtensionHook construction with start_enabled=False"""
        @six.add_metaclass(ExtensionHookPoint)
        class InitTestHook(ExtensionHook):
            def initialize(hook):
                self.fail('initialize() should not have been reached')

        hook = InitTestHook(self.extension, start_enabled=False)
        self.assertEqual(hook.hook_state, ExtensionHook.HOOK_STATE_DISABLED)
        self.assertFalse(hook.initialized)

    def test_registration(self):
        """Testing ExtensionHook registration"""
        extension_hook = TestExtensionHook(self.extension)

        self.assertEqual(self.extension, extension_hook.extension)
        self.assertIn(extension_hook, self.extension.hooks)
        self.assertIn(extension_hook, extension_hook.__class__.hooks)

    def test_shutdown(self):
        """Testing ExtensionHook.shutdown"""
        extension_hook = TestExtensionHook(self.extension)
        extension_hook.disable_hook()

        self.assertNotIn(extension_hook, extension_hook.__class__.hooks)



"""Unit tests for djblets.extensions.hooks.ExtensionHookPoint."""

from __future__ import unicode_literals

from django.utils import six
from mock import Mock

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.testing.testcases import TestCase


@six.add_metaclass(ExtensionHookPoint)
class TestExtensionHook(ExtensionHook):
    pass


class TestExtension(Extension):
    pass


class ExtensionHookPointTest(ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.ExtensionHookPoint."""

    def setUp(self):
        super(ExtensionHookPointTest, self).setUp()

        self.extension = self.setup_extension(TestExtension)
        self.extension_hook_class = TestExtensionHook
        self.dummy_hook = Mock()
        self.extension_hook_class.add_hook(self.dummy_hook)

    def test_extension_hook_class_gets_hooks(self):
        """Testing ExtensionHookPoint.hooks"""
        self.assertTrue(hasattr(self.extension_hook_class, "hooks"))

    def test_add_hook(self):
        """Testing ExtensionHookPoint.add_hook"""
        self.assertTrue(self.dummy_hook in self.extension_hook_class.hooks)

    def test_remove_hook(self):
        """Testing ExtensionHookPoint.remove_hook"""
        self.extension_hook_class.remove_hook(self.dummy_hook)
        self.assertTrue(self.dummy_hook not in self.extension_hook_class.hooks)



from __future__ import unicode_literals

from djblets.features import Feature
from djblets.features.errors import FeatureConflictError
from djblets.features.registry import FeaturesRegistry
from djblets.registries.errors import RegistrationError
from djblets.testing.testcases import TestCase


class DummyFeature(Feature):
    feature_id = 'djblets.tests.dummy-feature'
    name = 'Dummy Feature'
    summary = 'This is a dummy feature.'

    def initialize(self):
        self._feature_inited = True

    def shutdown(self):
        self._feature_inited = False


class FeaturesRegistryTests(TestCase):
    """Unit tests for djblets.features.registry.FeaturesRegistry."""

    def setUp(self):
        super(FeaturesRegistryTests, self).setUp()

        self.registry = FeaturesRegistry()

    def test_register_with_missing_id(self):
        """Testing FeaturesRegistry.register with missing ID"""
        class InvalidFeature(Feature):
            pass

        with self.assertRaises(RegistrationError):
            self.registry.register(InvalidFeature())

    def test_register_with_conflict(self):
        """Testing FeaturesRegistry.register with conflicting ID"""
        self.registry.register(DummyFeature(register=False))

        with self.assertRaises(FeatureConflictError):
            self.registry.register(DummyFeature(register=False))

    def test_register_calls_initialize(self):
        """Testing FeaturesRegistry.register initializes feature"""
        feature = DummyFeature(register=False)
        self.registry.register(feature)

        self.assertTrue(getattr(feature, '_feature_inited', False))

    def test_unregister_calls_shutdown(self):
        """Testing FeaturesRegistry.register shuts down feature"""
        feature = DummyFeature(register=False)

        self.registry.register(feature)
        self.assertTrue(getattr(feature, '_feature_inited', False))

        self.registry.unregister(feature)
        self.assertFalse(feature._feature_inited)

    def test_get_feature(self):
        """Testing FeaturesRegistry.get_feature"""
        feature = DummyFeature(register=False)

        self.registry.register(feature)
        self.assertEqual(self.registry.get_feature(feature.feature_id),
                         feature)

    def test_get_feature_with_invalid_id(self):
        """Testing FeaturesRegistry.get_feature with invalid ID"""
        self.assertEqual(self.registry.get_feature('bad-id'), None)



"""Tests for djblets.extensions.templatetags."""

from __future__ import unicode_literals

import os
import re
import uuid
from collections import OrderedDict

from django.core.urlresolvers import ResolverMatch
from django.http import HttpRequest
from django.template import Context, Template
from django.utils.safestring import mark_safe
from kgb import SpyAgency

from djblets.extensions.extension import Extension, ExtensionInfo, JSExtension
from djblets.extensions.hooks import TemplateHook
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.testing.testcases import TestCase
from pipeline.conf import settings as pipeline_settings


class TestJSExtension(JSExtension):
    model_class = 'FooNew'

    def get_model_data(self, request, **kwargs):
        return {'test': 'new'}


class TestJSExtensionDeprecated(JSExtension):
    model_class = 'FooOld'

    def get_model_data(self):
        return {'test': 'old'}


class TemplateTagTests(SpyAgency, ExtensionTestsMixin, TestCase):
    """Tests for djblets.extensions.templatetags."""

    class TestExtension(Extension):
        # While normally these would be plain dictionaries, we want to ensure
        # order for testing, so we'll be doing this a bit differently.
        css_bundles = OrderedDict()
        css_bundles['default'] = {
            'source_filenames': (
                'css/default-test.css',
            ),
        }
        css_bundles['optional'] = {
            'source_filenames': (
                'css/optional-test.css',
            ),
            'apply_to': ['foo'],
        }

        js_bundles = OrderedDict()
        js_bundles['default'] = {
            'source_filenames': (
                'js/default-test.js',
            ),
        }
        js_bundles['optional'] = {
            'source_filenames': (
                'js/optional-test.js',
            ),
            'apply_to': ['foo'],
        }

        js_extensions = [TestJSExtension, TestJSExtensionDeprecated]

    def setUp(self):
        def _has_resource(self, path):
            return path == 'static'

        def _extract_resource(self, path):
            if path == 'static':
                return os.path.abspath(os.path.join(__file__, '..', 'static'))

            return None

        super(TemplateTagTests, self).setUp()

        self.key = uuid.uuid4()

        self.spy_on(ExtensionInfo.has_resource,
                    owner=ExtensionInfo,
                    call_fake=_has_resource)
        self.spy_on(ExtensionInfo.extract_resource,
                    owner=ExtensionInfo,
                    call_fake=_extract_resource)

        self.extension = self.setup_extension(self.TestExtension,
                                              manager_key=self.key)
        self.request = HttpRequest()

        self._old_pipeline_enabled = pipeline_settings.PIPELINE_ENABLED

    def tearDown(self):
        super(TemplateTagTests, self).tearDown()

        pipeline_settings.PIPELINE_ENABLED = self._old_pipeline_enabled

    def test_template_hook_point_tag(self):
        """Testing template_hook_point template tag"""
        class DummyTemplateHook(TemplateHook):
            """A template hook where the content is specified."""

            def __init__(self, content, *args, **kwargs):
                super(DummyTemplateHook, self).__init__(*args, **kwargs)
                self._content = content

            def render_to_string(self, request, context):
                return self._content

        t = Template('{% load djblets_extensions %}'
                     '{% template_hook_point "test-hook" %}')

        DummyTemplateHook(content=mark_safe('<strong>foo</strong>\n'),
                          extension=self.extension,
                          name='test-hook')
        DummyTemplateHook(content='<bar>\n',
                          extension=self.extension,
                          name='test-hook')

        self.assertEqual(
            t.render(Context({
                'request': self.request,
            })),
            '<strong>foo</strong>\n'
            '&lt;bar&gt;\n')

    def test_ext_static_tag(self):
        """Testing ext_static template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% ext_static ext "css/default-test.css" %}')

        self.assertEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '/ext/djblets.extensions.tests.test_templatetags.TestExtension/'
            'css/default-test.dad0c9b31e59.css')

    def test_ext_css_bundle_tag_with_pipeline_enabled(self):
        """Testing ext_css_bundle template tag with PIPELINE_ENABLED=True"""
        pipeline_settings.PIPELINE_ENABLED = True

        t = Template('{% load djblets_extensions %}'
                     '{% ext_css_bundle ext "default" %}')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/css/default.min.dad0c9b31e59.css" '
            'rel="stylesheet" type="text/css" />')

    def test_ext_css_bundle_tag_with_pipeline_disabled(self):
        """Testing ext_css_bundle template tag with PIPELINE_ENABLED=False"""
        pipeline_settings.PIPELINE_ENABLED = False

        t = Template('{% load djblets_extensions %}'
                     '{% ext_css_bundle ext "default" %}')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '<link href="/ext/djblets.extensions.tests.test_templatetags.'
            'TestExtension/css/default-test.dad0c9b31e59.css" '
            'rel="stylesheet" type="text/css" />')

    def test_ext_js_bundle_tag_with_pipeline_enabled(self):
        """Testing ext_js_bundle template tag with PIPELINE_ENABLED=True"""
        pipeline_settings.PIPELINE_ENABLED = True

        t = Template('{% load djblets_extensions %}'
                     '{% ext_js_bundle ext "default" %}')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetag'
            's.TestExtension/js/default.min.dad0c9b31e59.js" '
            'charset="utf-8"></script>')

    def test_ext_js_bundle_tag_with_pipeline_disabled(self):
        """Testing ext_js_bundle template tag with PIPELINE_ENABLED=False"""
        pipeline_settings.PIPELINE_ENABLED = False

        t = Template('{% load djblets_extensions %}'
                     '{% ext_js_bundle ext "default" %}')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'request': self.request,
            })),
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetag'
            's.TestExtension/js/default-test.dad0c9b31e59.js" '
            'charset="utf-8"></script>')

    def test_load_extensions_css_tag_with_pipline_enabled(self):
        """Testing load_extensions_css template tag with PIPELINE_ENABLED=True
        """
        pipeline_settings.PIPELINE_ENABLED = True

        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_css manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.key,
                'request': self.request,
            })),
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/css/default.min.dad0c9b31e59.css" '
            'rel="stylesheet" type="text/css" />'
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/css/optional.min.dad0c9b31e59.css" '
            'rel="stylesheet" type="text/css" />')

    def test_load_extensions_css_tag_with_pipline_disabled(self):
        """Testing load_extensions_css template tag with PIPELINE_ENABLED=False
        """
        pipeline_settings.PIPELINE_ENABLED = False

        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_css manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.key,
                'request': self.request,
            })),
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/css/default-test.dad0c9b31e59.css" '
            'rel="stylesheet" type="text/css" />'
            '<link href="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/css/optional-test.dad0c9b31e59.css" '
            'rel="stylesheet" type="text/css" />')

    def test_load_extensions_js_tag_with_pipeline_enabled(self):
        """Testing load_extensions_js template tag with PIPELINE_ENABLED=True
        """
        pipeline_settings.PIPELINE_ENABLED = True

        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_js manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.key,
                'request': self.request,
            })),
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/js/default.min.dad0c9b31e59.js" '
            'charset="utf-8"></script>'
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/js/optional.min.dad0c9b31e59.js" '
            'charset="utf-8"></script>')

    def test_load_extensions_js_tag_with_pipeline_disabled(self):
        """Testing load_extensions_js template tag with PIPELINE_ENABLED=False
        """
        pipeline_settings.PIPELINE_ENABLED = False

        t = Template('{% load djblets_extensions %}'
                     '{% load_extensions_js manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        self.assertHTMLEqual(
            t.render(Context({
                'ext': self.extension,
                'manager_id': self.key,
                'request': self.request,
            })),
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/js/default-test.dad0c9b31e59.js" '
            'charset="utf-8"></script>'
            '<script type="text/javascript" '
            'src="/ext/djblets.extensions.tests.test_templatetags'
            '.TestExtension/js/optional-test.dad0c9b31e59.js" '
            'charset="utf-8"></script>')

    def test_init_js_extensions(self):
        """Testing init_js_extensions template tag"""
        t = Template('{% load djblets_extensions %}'
                     '{% init_js_extensions manager_id %}')

        self.request.resolver_match = ResolverMatch(None, None, None, 'foo')

        content = t.render(Context({
            'ext': self.extension,
            'manager_id': self.key,
            'request': self.request,
        }))

        self.assertIsNotNone(re.search(
            r'new FooNew\({\s+"test": "new",',
            content))

        self.assertIsNotNone(re.search(
            r'new FooOld\({\s+"test": "old",',
            content))
